"""
This example follows the original Cargo Shipping example
that figures in the DDD book, as coded in the "DDD Sample"
project. As it says on the project website, "This project is a
joint effort by Eric Evans' company Domain Language and
the Swedish software consulting company Citerus.

  -  http://dddsample.sourceforge.net/

"One of the most requested aids to coming up to speed on DDD
has been a running example application. Starting from a simple
set of functions and a model based on the cargo example used
in Eric Evans' book, we have built a running application with
which to demonstrate a practical implementation of the building
block patterns as well as illustrate the impact of aggregates
and bounded contexts."

"""
from datetime import datetime, timedelta
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple, Type, TypeVar, Union
from unittest import TestCase
from uuid import UUID

from eventsourcing.application.process import ProcessApplication
from eventsourcing.application.sqlalchemy import SQLAlchemyApplication
from eventsourcing.domain.model.aggregate import (
    AggregateRoot,
    TAggregate,
    TAggregateEvent,
)
from eventsourcing.system.definition import System
from eventsourcing.system.runner import InProcessRunner, SingleThreadedRunner


# Locations in the world.
class Location(Enum):
    HAMBURG = "HAMBURG"
    HONGKONG = "HONGKONG"
    NEWYORK = "NEWYORK"
    STOCKHOLM = "STOCKHOLM"
    TOKYO = "TOKYO"

    NLRTM = "NLRTM"
    USDAL = "USDAL"
    AUMEL = "AUMEL"


# Leg of an Itinerary.
class Leg(object):
    def __init__(self, origin: str, destination: str, voyage_number: str):
        self.origin: str = origin
        self.destination: str = destination
        self.voyage_number: str = voyage_number


# Itinerary.
class Itinerary(object):
    def __init__(self, origin: str, destination: str, legs: List[Leg]):
        self.origin = origin
        self.destination = destination
        self.legs = legs


# Some routes from one location to another.
REGISTERED_ROUTES = {
    ("HONGKONG", "STOCKHOLM"): [
        Itinerary(
            origin="HONGKONG",
            destination="STOCKHOLM",
            legs=[
                Leg(origin="HONGKONG", destination="NEWYORK", voyage_number="V1"),
                Leg(origin="NEWYORK", destination="STOCKHOLM", voyage_number="V2"),
            ],
        )
    ],
    ("TOKYO", "STOCKHOLM"): [
        Itinerary(
            origin="TOKYO",
            destination="STOCKHOLM",
            legs=[
                Leg(origin="TOKYO", destination="HAMBURG", voyage_number="V3"),
                Leg(origin="HAMBURG", destination="STOCKHOLM", voyage_number="V4"),
            ],
        )
    ],
}


# Handling activities.
class HandlingActivity(Enum):
    RECEIVE = "RECEIVE"
    LOAD = "LOAD"
    UNLOAD = "UNLOAD"
    CLAIM = "CLAIM"


# Custom static types.
NextExpectedActivity = Optional[
    Union[Tuple[HandlingActivity, Location], Tuple[HandlingActivity, Location, str]]
]
CargoDetails = Dict[str, Optional[Union[str, bool, datetime, Tuple]]]
LegDetails = Dict[str, str]
ItineraryDetails = Dict[str, Union[str, List[LegDetails]]]

# Type variable for Cargo aggregate class.
T_cargo = TypeVar("T_cargo", bound="Cargo")


# Custom aggregate root class.
class Aggregate(AggregateRoot):
    __subclassevents__ = True


# The Cargo aggregate is an event sourced domain model aggregate that
# specifies the routing from origin to destination, and can track what
# happens to the cargo after it has been booked.


class Cargo(Aggregate):
    @classmethod
    def new_booking(
        cls: Type[T_cargo],
        origin: Location,
        destination: Location,
        arrival_deadline: datetime,
    ) -> T_cargo:
        assert issubclass(cls, Cargo)  # For PyCharm navigation.
        return cls.__create__(
            origin=origin, destination=destination, arrival_deadline=arrival_deadline
        )

    def __init__(
        self,
        origin: Location,
        destination: Location,
        arrival_deadline: datetime,
        **kwargs: Any
    ) -> None:
        super().__init__(**kwargs)
        self._origin: Location = origin
        self._destination: Location = destination
        self._arrival_deadline: datetime = arrival_deadline
        self._transport_status: str = "NOT_RECEIVED"
        self._routing_status: str = "NOT_ROUTED"
        self._is_misdirected: bool = False
        self._estimated_time_of_arrival: Optional[datetime] = None
        self._next_expected_activity: NextExpectedActivity = None
        self._route: Optional[Itinerary] = None
        self._last_known_location: Optional[Location] = None
        self._current_voyage_number: Optional[str] = None

    @property
    def origin(self) -> Location:
        return self._origin

    @property
    def destination(self) -> Location:
        return self._destination

    @property
    def arrival_deadline(self) -> datetime:
        return self._arrival_deadline

    @property
    def transport_status(self) -> str:
        return self._transport_status

    @property
    def routing_status(self) -> str:
        return self._routing_status

    @property
    def is_misdirected(self) -> bool:
        return self._is_misdirected

    @property
    def estimated_time_of_arrival(self) -> Optional[datetime]:
        return self._estimated_time_of_arrival

    @property
    def next_expected_activity(self) -> Optional[Tuple]:
        return self._next_expected_activity

    @property
    def route(self) -> Optional[Itinerary]:
        return self._route

    @property
    def last_known_location(self) -> Optional[Location]:
        return self._last_known_location

    @property
    def current_voyage_number(self) -> Optional[str]:
        return self._current_voyage_number

    class Event(Aggregate.Event):
        pass

    def change_destination(self, destination: Location) -> None:
        self.__trigger_event__(self.DestinationChanged, destination=destination)

    class DestinationChanged(Event):
        def mutate(self, obj: "Cargo") -> None:
            obj._destination = self.destination

        @property
        def destination(self) -> Location:
            return self.__dict__["destination"]

    def assign_route(self, itinerary: Itinerary) -> None:
        self.__trigger_event__(self.RouteAssigned, route=itinerary)

    class RouteAssigned(Event):
        def mutate(self, obj: "Cargo") -> None:
            obj._route = self.route
            obj._routing_status = "ROUTED"
            obj._estimated_time_of_arrival = datetime.now() + timedelta(weeks=1)
            obj._next_expected_activity = (HandlingActivity.RECEIVE, obj.origin)
            obj._is_misdirected = False

        @property
        def route(self) -> Itinerary:
            return self.__dict__["route"]

    def register_handling_event(
        self,
        tracking_id: UUID,
        voyage_number: Optional[str],
        location: Location,
        handling_activity: HandlingActivity,
    ) -> None:
        self.__trigger_event__(
            self.HandlingEventRegistered,
            tracking_id=tracking_id,
            voyage_number=voyage_number,
            location=location,
            handling_activity=handling_activity,
        )

    class HandlingEventRegistered(Event):
        def mutate(self, obj: "Cargo") -> None:
            assert obj.route is not None
            if self.handling_activity == HandlingActivity.RECEIVE:
                obj._transport_status = "IN_PORT"
                obj._last_known_location = self.location
                obj._next_expected_activity = (
                    HandlingActivity.LOAD,
                    self.location,
                    obj.route.legs[0].voyage_number,
                )
            elif self.handling_activity == HandlingActivity.LOAD:
                obj._transport_status = "ONBOARD_CARRIER"
                obj._current_voyage_number = self.voyage_number
                for leg in obj.route.legs:
                    if leg.origin == self.location.value:
                        if leg.voyage_number == self.voyage_number:
                            obj._next_expected_activity = (
                                HandlingActivity.UNLOAD,
                                Location[leg.destination],
                                self.voyage_number,
                            )
                            break
                else:
                    raise Exception(
                        "Can't find leg with origin={} and "
                        "voyage_number={}".format(self.location, self.voyage_number)
                    )

            elif self.handling_activity == HandlingActivity.UNLOAD:
                obj._current_voyage_number = None
                obj._last_known_location = self.location
                obj._transport_status = "IN_PORT"
                if self.location == obj.destination:
                    obj._next_expected_activity = (
                        HandlingActivity.CLAIM,
                        self.location,
                    )
                elif self.location.value in [leg.destination for leg in obj.route.legs]:
                    for i, leg in enumerate(obj.route.legs):
                        if leg.voyage_number == self.voyage_number:
                            next_leg: Leg = obj.route.legs[i + 1]
                            assert Location[next_leg.origin] == self.location
                            obj._next_expected_activity = (
                                HandlingActivity.LOAD,
                                self.location,
                                next_leg.voyage_number,
                            )
                            break
                else:
                    obj._is_misdirected = True
                    obj._next_expected_activity = None

            elif self.handling_activity == HandlingActivity.CLAIM:
                obj._next_expected_activity = None
                obj._transport_status = "CLAIMED"

            else:
                raise Exception(
                    "Unsupported handling event: {}".format(self.handling_activity)
                )

        @property
        def voyage_number(self) -> str:
            return self.__dict__["voyage_number"]

        @property
        def location(self) -> Location:
            return self.__dict__["location"]

        @property
        def handling_activity(self) -> str:
            return self.__dict__["handling_activity"]


# Cargo aggregates exist within an application, which
# provides "application service" methods for clients.
class BookingApplication(ProcessApplication[TAggregate, TAggregateEvent]):
    persist_event_type = Cargo.Event

    @staticmethod
    def book_new_cargo(
        origin: Location, destination: Location, arrival_deadline: datetime
    ) -> UUID:
        cargo = Cargo.new_booking(origin, destination, arrival_deadline)
        cargo.__save__()
        return cargo.id

    def change_destination(self, tracking_id: UUID, destination: Location) -> None:
        cargo = self.get_cargo(tracking_id)
        cargo.change_destination(destination)
        cargo.__save__()

    def request_possible_routes_for_cargo(self, tracking_id: UUID) -> List[Itinerary]:
        cargo = self.get_cargo(tracking_id)
        from_location = (cargo.last_known_location or cargo.origin).value
        to_location = cargo.destination.value
        try:
            possible_routes = REGISTERED_ROUTES[(from_location, to_location)]
        except KeyError:
            raise Exception(
                "Can't find routes from {} to {}".format(from_location, to_location)
            )

        return possible_routes

    def assign_route(self, tracking_id: UUID, itinerary: Itinerary) -> None:
        cargo = self.get_cargo(tracking_id)
        cargo.assign_route(itinerary)
        cargo.__save__()

    def register_handling_event(
        self,
        tracking_id: UUID,
        voyage_number: Optional[str],
        location: Location,
        handing_activity: HandlingActivity,
    ) -> None:
        cargo = self.get_cargo(tracking_id)
        cargo.register_handling_event(
            tracking_id, voyage_number, location, handing_activity
        )
        cargo.__save__()

    def get_cargo(self, tracking_id: UUID) -> Cargo:
        cargo = self.repository.get_instance_of(Cargo, tracking_id)
        if cargo is None:
            raise Exception("Cargo not found: {}".format(tracking_id))
        else:
            return cargo


# The application services are presented in a client interface that
# deals with simple types of object (str, bool, datetime).
class LocalClient(object):
    def __init__(self, runner: InProcessRunner):
        self.runner: InProcessRunner = runner
        self.booking_application = self.runner.get(BookingApplication)

    def book_new_cargo(
        self, origin: str, destination: str, arrival_deadline: datetime
    ) -> str:
        tracking_id = self.booking_application.book_new_cargo(
            Location[origin], Location[destination], arrival_deadline
        )
        return str(tracking_id)

    def get_cargo_details(self, tracking_id: str) -> CargoDetails:
        cargo = self.booking_application.get_cargo(UUID(tracking_id))

        # Present 'next_expected_activity'.
        next_expected_activity: Optional[Union[Tuple[Any, Any], Tuple[Any, Any, Any]]]
        if cargo.next_expected_activity is None:
            next_expected_activity = None
        elif len(cargo.next_expected_activity) == 2:
            next_expected_activity = (
                cargo.next_expected_activity[0].value,
                cargo.next_expected_activity[1].value,
            )
        elif len(cargo.next_expected_activity) == 3:
            next_expected_activity = (
                cargo.next_expected_activity[0].value,
                cargo.next_expected_activity[1].value,
                cargo.next_expected_activity[2],
            )
        else:
            raise Exception(
                "Invalid next expected activity: {}".format(
                    cargo.next_expected_activity
                )
            )

        # Present 'last_known_location'.
        if cargo.last_known_location is None:
            last_known_location = None
        else:
            last_known_location = cargo.last_known_location.value

        # Present the cargo details.
        return {
            "id": str(cargo.id),
            "origin": cargo.origin.value,
            "destination": cargo.destination.value,
            "arrival_deadline": cargo.arrival_deadline,
            "transport_status": cargo.transport_status,
            "routing_status": cargo.routing_status,
            "is_misdirected": cargo.is_misdirected,
            "estimated_time_of_arrival": cargo.estimated_time_of_arrival,
            "next_expected_activity": next_expected_activity,
            "last_known_location": last_known_location,
            "current_voyage_number": cargo.current_voyage_number,
        }

    def change_destination(self, tracking_id: str, destination: str) -> None:
        self.booking_application.change_destination(
            UUID(tracking_id), Location[destination]
        )

    def request_possible_routes_for_cargo(self, tracking_id: str) -> List[dict]:
        routes = self.booking_application.request_possible_routes_for_cargo(
            UUID(tracking_id)
        )
        return [self.dict_from_itinerary(route) for route in routes]

    def dict_from_itinerary(self, itinerary: Itinerary) -> ItineraryDetails:
        legs_details = []
        for leg in itinerary.legs:
            leg_details: LegDetails = {
                "origin": leg.origin,
                "destination": leg.destination,
                "voyage_number": leg.voyage_number,
            }
            legs_details.append(leg_details)
        route_details: ItineraryDetails = {
            "origin": itinerary.origin,
            "destination": itinerary.destination,
            "legs": legs_details,
        }
        return route_details

    def assign_route(self, tracking_id: str, route_details: ItineraryDetails) -> None:
        routes = self.booking_application.request_possible_routes_for_cargo(
            UUID(tracking_id)
        )
        for route in routes:
            if route_details == self.dict_from_itinerary(route):
                self.booking_application.assign_route(UUID(tracking_id), route)

    def register_handling_event(
        self,
        tracking_id: str,
        voyage_number: Optional[str],
        location: str,
        handling_activity: str,
    ) -> None:
        self.booking_application.register_handling_event(
            UUID(tracking_id),
            voyage_number,
            Location[location],
            HandlingActivity[handling_activity],
        )


# Stub function that picks an itineraries from a list of possible itineraries.
def select_preferred_itinerary(itineraries: List[ItineraryDetails]) -> ItineraryDetails:
    return itineraries[0]


class TestDDDCargoShippingExample(TestCase):
    def setUp(self) -> None:
        self.runner = SingleThreadedRunner(
            system=System(BookingApplication),
            infrastructure_class=SQLAlchemyApplication,
            setup_tables=True,
        )
        self.runner.start()
        self.client = LocalClient(self.runner)

    def tearDown(self) -> None:
        self.runner.close()

    def test_admin_can_book_new_cargo(self) -> None:
        arrival_deadline = datetime.now() + timedelta(weeks=3)

        cargo_id = self.client.book_new_cargo(
            origin="NLRTM", destination="USDAL", arrival_deadline=arrival_deadline
        )

        cargo_details = self.client.get_cargo_details(cargo_id)
        self.assertTrue(cargo_details["id"])
        self.assertEqual(cargo_details["origin"], "NLRTM")
        self.assertEqual(cargo_details["destination"], "USDAL")

        self.client.change_destination(cargo_id, destination="AUMEL")
        cargo_details = self.client.get_cargo_details(cargo_id)
        self.assertEqual(cargo_details["destination"], "AUMEL")
        self.assertEqual(cargo_details["arrival_deadline"], arrival_deadline)

    def test_scenario_cargo_from_hongkong_to_stockholm(self) -> None:
        # Test setup: A cargo should be shipped from Hongkong to Stockholm,
        # and it should arrive in no more than two weeks.
        origin = "HONGKONG"
        destination = "STOCKHOLM"
        arrival_deadline = datetime.now() + timedelta(weeks=2)

        # Use case 1: booking.

        # A new cargo is booked, and the unique tracking id is assigned to the cargo.
        tracking_id = self.client.book_new_cargo(origin, destination, arrival_deadline)

        # The tracking id can be used to lookup the cargo in the repository.
        # Important: The cargo, and thus the domain model, is responsible for
        # determining the status of the cargo, whether it is on the right track
        # or not and so on. This is core domain logic. Tracking the cargo basically
        # amounts to presenting information extracted from the cargo aggregate in a
        # suitable way.
        cargo_details = self.client.get_cargo_details(tracking_id)
        self.assertEqual(cargo_details["transport_status"], "NOT_RECEIVED")
        self.assertEqual(cargo_details["routing_status"], "NOT_ROUTED")
        self.assertEqual(cargo_details["is_misdirected"], False)
        self.assertEqual(cargo_details["estimated_time_of_arrival"], None)
        self.assertEqual(cargo_details["next_expected_activity"], None)

        # Use case 2: routing.
        #
        # A number of possible routes for this cargo is requested and may be
        # presented to the customer in some way for him/her to choose from.
        # Selection could be affected by things like price and time of delivery,
        # but this test simply uses an arbitrary selection to mimic that process.
        routes_details = self.client.request_possible_routes_for_cargo(tracking_id)
        route_details = select_preferred_itinerary(routes_details)

        # The cargo is then assigned to the selected route, described by an itinerary.
        self.client.assign_route(tracking_id, route_details)

        cargo_details = self.client.get_cargo_details(tracking_id)
        self.assertEqual(cargo_details["transport_status"], "NOT_RECEIVED")
        self.assertEqual(cargo_details["routing_status"], "ROUTED")
        self.assertEqual(cargo_details["is_misdirected"], False)
        self.assertTrue(cargo_details["estimated_time_of_arrival"])
        self.assertEqual(
            cargo_details["next_expected_activity"], ("RECEIVE", "HONGKONG")
        )

        # Use case 3: handling

        # A handling event registration attempt will be formed from parsing
        # the data coming in as a handling report either via the web service
        # interface or as an uploaded CSV file. The handling event factory
        # tries to create a HandlingEvent from the attempt, and if the factory
        # decides that this is a plausible handling event, it is stored.
        # If the attempt is invalid, for example if no cargo exists for the
        # specfied tracking id, the attempt is rejected.
        #
        # Handling begins: cargo is received in Hongkong.
        self.client.register_handling_event(tracking_id, None, "HONGKONG", "RECEIVE")
        cargo_details = self.client.get_cargo_details(tracking_id)
        self.assertEqual(cargo_details["transport_status"], "IN_PORT")
        self.assertEqual(cargo_details["last_known_location"], "HONGKONG")
        self.assertEqual(
            cargo_details["next_expected_activity"], ("LOAD", "HONGKONG", "V1")
        )

        # Load onto voyage V1.
        self.client.register_handling_event(tracking_id, "V1", "HONGKONG", "LOAD")
        cargo_details = self.client.get_cargo_details(tracking_id)
        self.assertEqual(cargo_details["current_voyage_number"], "V1")
        self.assertEqual(cargo_details["last_known_location"], "HONGKONG")
        self.assertEqual(cargo_details["transport_status"], "ONBOARD_CARRIER")
        self.assertEqual(
            cargo_details["next_expected_activity"], ("UNLOAD", "NEWYORK", "V1")
        )

        # Incorrectly unload in Tokyo.
        self.client.register_handling_event(tracking_id, "V1", "TOKYO", "UNLOAD")
        cargo_details = self.client.get_cargo_details(tracking_id)
        self.assertEqual(cargo_details["current_voyage_number"], None)
        self.assertEqual(cargo_details["last_known_location"], "TOKYO")
        self.assertEqual(cargo_details["transport_status"], "IN_PORT")
        self.assertEqual(cargo_details["is_misdirected"], True)
        self.assertEqual(cargo_details["next_expected_activity"], None)

        # Reroute.
        routes_details = self.client.request_possible_routes_for_cargo(tracking_id)
        route_details = select_preferred_itinerary(routes_details)
        self.client.assign_route(tracking_id, route_details)

        # Load in Tokyo.
        self.client.register_handling_event(tracking_id, "V3", "TOKYO", "LOAD")
        cargo_details = self.client.get_cargo_details(tracking_id)
        self.assertEqual(cargo_details["current_voyage_number"], "V3")
        self.assertEqual(cargo_details["last_known_location"], "TOKYO")
        self.assertEqual(cargo_details["transport_status"], "ONBOARD_CARRIER")
        self.assertEqual(cargo_details["is_misdirected"], False)
        self.assertEqual(
            cargo_details["next_expected_activity"], ("UNLOAD", "HAMBURG", "V3")
        )

        # Unload in Hamburg.
        self.client.register_handling_event(tracking_id, "V3", "HAMBURG", "UNLOAD")
        cargo_details = self.client.get_cargo_details(tracking_id)
        self.assertEqual(cargo_details["current_voyage_number"], None)
        self.assertEqual(cargo_details["last_known_location"], "HAMBURG")
        self.assertEqual(cargo_details["transport_status"], "IN_PORT")
        self.assertEqual(cargo_details["is_misdirected"], False)
        self.assertEqual(
            cargo_details["next_expected_activity"], ("LOAD", "HAMBURG", "V4")
        )

        # Load in Hamburg
        self.client.register_handling_event(tracking_id, "V4", "HAMBURG", "LOAD")
        cargo_details = self.client.get_cargo_details(tracking_id)
        self.assertEqual(cargo_details["current_voyage_number"], "V4")
        self.assertEqual(cargo_details["last_known_location"], "HAMBURG")
        self.assertEqual(cargo_details["transport_status"], "ONBOARD_CARRIER")
        self.assertEqual(cargo_details["is_misdirected"], False)
        self.assertEqual(
            cargo_details["next_expected_activity"], ("UNLOAD", "STOCKHOLM", "V4")
        )

        # Unload in Stockholm
        self.client.register_handling_event(tracking_id, "V4", "STOCKHOLM", "UNLOAD")
        cargo_details = self.client.get_cargo_details(tracking_id)
        self.assertEqual(cargo_details["current_voyage_number"], None)
        self.assertEqual(cargo_details["last_known_location"], "STOCKHOLM")
        self.assertEqual(cargo_details["transport_status"], "IN_PORT")
        self.assertEqual(cargo_details["is_misdirected"], False)
        self.assertEqual(
            cargo_details["next_expected_activity"], ("CLAIM", "STOCKHOLM")
        )

        # Finally, cargo is claimed in Stockholm.
        self.client.register_handling_event(tracking_id, None, "STOCKHOLM", "CLAIM")
        cargo_details = self.client.get_cargo_details(tracking_id)
        self.assertEqual(cargo_details["current_voyage_number"], None)
        self.assertEqual(cargo_details["last_known_location"], "STOCKHOLM")
        self.assertEqual(cargo_details["transport_status"], "CLAIMED")
        self.assertEqual(cargo_details["is_misdirected"], False)
        self.assertEqual(cargo_details["next_expected_activity"], None)



import multiprocessing
from multiprocessing import Manager
from queue import Empty, Queue
from time import sleep
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Sequence, Tuple, Type

from eventsourcing.application.notificationlog import RecordManagerNotificationLog
from eventsourcing.application.popo import PopoApplication
from eventsourcing.application.process import ProcessApplication, PromptToQuit
from eventsourcing.application.simple import (
    ApplicationWithConcreteInfrastructure,
    Prompt,
    PromptToPull,
    is_prompt_to_pull,
)
from eventsourcing.domain.model.decorators import retry
from eventsourcing.domain.model.events import subscribe, unsubscribe
from eventsourcing.exceptions import (
    CausalDependencyFailed,
    OperationalError,
    ProgrammingError,
    RecordConflictError,
)
from eventsourcing.infrastructure.base import DEFAULT_PIPELINE_ID
from eventsourcing.system.definition import AbstractSystemRunner, System
from eventsourcing.system.runner import DEFAULT_POLL_INTERVAL, PromptOutbox


class MultiprocessRunner(AbstractSystemRunner):
    def __init__(
        self,
        system: System,
        pipeline_ids: Sequence[int] = (DEFAULT_PIPELINE_ID,),
        poll_interval: Optional[int] = None,
        setup_tables: bool = False,
        sleep_for_setup_tables: int = 0,
        **kwargs: Any
    ):
        super(MultiprocessRunner, self).__init__(system=system, **kwargs)
        self.pipeline_ids = pipeline_ids
        self.poll_interval = poll_interval or DEFAULT_POLL_INTERVAL
        assert isinstance(system, System)
        self.setup_tables = setup_tables or system.setup_tables
        self.sleep_for_setup_tables = sleep_for_setup_tables
        self.os_processes: List[OperatingSystemProcess] = []

    def start(self) -> None:
        self.os_processes = []

        self.manager = Manager()

        if TYPE_CHECKING:
            self.inboxes: Dict[Tuple[int, str], Queue[Prompt]]
            self.outboxes: Dict[Tuple[int, str], PromptOutbox[Tuple[int, str]]]
        self.inboxes = {}
        self.outboxes = {}

        # Setup queues.
        for pipeline_id in self.pipeline_ids:
            for process_name, upstream_names in self.system.upstream_names.items():
                inbox_id = (pipeline_id, process_name.lower())
                if inbox_id not in self.inboxes:
                    self.inboxes[inbox_id] = self.manager.Queue()
                for upstream_class_name in upstream_names:
                    outbox_id = (pipeline_id, upstream_class_name.lower())
                    if outbox_id not in self.outboxes:
                        self.outboxes[outbox_id] = PromptOutbox()
                    if inbox_id not in self.outboxes[outbox_id].downstream_inboxes:
                        self.outboxes[outbox_id].downstream_inboxes[
                            inbox_id
                        ] = self.inboxes[inbox_id]

        # Check we have the infrastructure classes we need.
        for process_class in self.system.process_classes.values():
            if not isinstance(process_class, ApplicationWithConcreteInfrastructure):
                if not self.infrastructure_class:
                    raise ProgrammingError("infrastructure_class is not set")
                elif issubclass(self.infrastructure_class, PopoApplication):
                    raise ProgrammingError("Can't use %s with %s" % (
                        type(self), self.infrastructure_class
                    ))
                elif not issubclass(
                    self.infrastructure_class, ApplicationWithConcreteInfrastructure
                ):
                    raise ProgrammingError(
                        "infrastructure_class is not a subclass of {}".format(
                            ApplicationWithConcreteInfrastructure
                        )
                    )

        # Subscribe to broadcast prompts published by a process
        # application in the parent operating system process.
        subscribe(handler=self.broadcast_prompt, predicate=is_prompt_to_pull)

        # Start operating system process.
        expect_tables_exist = False
        for pipeline_id in self.pipeline_ids:
            for process_name, upstream_names in self.system.upstream_names.items():
                process_class = self.system.process_classes[process_name]
                inbox = self.inboxes[(pipeline_id, process_name.lower())]
                outbox = self.outboxes.get((pipeline_id, process_name.lower()))
                os_process = OperatingSystemProcess(
                    application_process_class=process_class,
                    infrastructure_class=self.infrastructure_class,
                    upstream_names=upstream_names,
                    poll_interval=self.poll_interval,
                    pipeline_id=pipeline_id,
                    setup_tables=self.setup_tables,
                    inbox=inbox,
                    outbox=outbox,
                )
                os_process.daemon = True
                os_process.start()
                self.os_processes.append(os_process)
                if self.setup_tables and not expect_tables_exist:
                    # Avoid conflicts when creating tables.
                    sleep(self.sleep_for_setup_tables)
                    expect_tables_exist = True

        # Construct process applications in local process.
        for process_class in self.system.process_classes.values():
            self.get(process_class)

    def broadcast_prompt(self, prompt: PromptToPull) -> None:
        outbox_id = (prompt.pipeline_id, prompt.process_name)
        outbox = self.outboxes.get(outbox_id)
        if outbox:
            outbox.put(prompt)

    def close(self) -> None:
        super(MultiprocessRunner, self).close()

        unsubscribe(handler=self.broadcast_prompt, predicate=is_prompt_to_pull)

        for os_process in self.os_processes:
            os_process.inbox.put(PromptToQuit())

        for os_process in self.os_processes:
            os_process.join(timeout=10)

        for os_process in self.os_processes:
            if os_process.is_alive():
                os_process.terminate()

        self.os_processes.clear()


class OperatingSystemProcess(multiprocessing.Process):
    def __init__(
        self,
        application_process_class: Type[ProcessApplication],
        infrastructure_class: Type[ApplicationWithConcreteInfrastructure],
        upstream_names: List[str],
        inbox: Queue,
        outbox: Optional[PromptOutbox[Tuple[int, str]]] = None,
        pipeline_id: int = DEFAULT_PIPELINE_ID,
        poll_interval: int = DEFAULT_POLL_INTERVAL,
        setup_tables: bool = False,
        *args: Any,
        **kwargs: Any
    ):
        super(OperatingSystemProcess, self).__init__(*args, **kwargs)
        self.application_process_class = application_process_class
        self.infrastructure_class = infrastructure_class
        self.upstream_names = upstream_names
        self.daemon = True
        self.pipeline_id = pipeline_id
        self.poll_interval = poll_interval
        self.inbox = inbox
        self.outbox = outbox
        self.setup_tables = setup_tables

    def run(self) -> None:
        # Construct process application class.
        process_class = self.application_process_class
        if not isinstance(process_class, ApplicationWithConcreteInfrastructure):
            if self.infrastructure_class:
                process_class = process_class.mixin(self.infrastructure_class)
            else:
                raise ProgrammingError("infrastructure_class is not set")

        # Construct process application object.
        self.process: ProcessApplication = process_class(
            pipeline_id=self.pipeline_id, setup_table=self.setup_tables
        )

        # Follow upstream notification logs.
        for upstream_name in self.upstream_names:

            # Obtain a notification log object (local or remote) for the upstream
            # process.
            if upstream_name == self.process.name:
                # Upstream is this process's application,
                # so use own notification log.
                notification_log = self.process.notification_log
            else:
                # For a different application, we need to construct a notification
                # log with a record manager that has the upstream application ID.
                # Currently assumes all applications are using the same database
                # and record manager class. If it wasn't the same database,we would
                # to use a remote notification log, and upstream would need to provide
                # an API from which we can pull. It's not unreasonable to have a fixed
                # number of application processes connecting to the same database.
                record_manager = self.process.event_store.record_manager

                notification_log = RecordManagerNotificationLog(
                    record_manager=record_manager.clone(
                        application_name=upstream_name,
                        # Todo: Check if setting pipeline_id is necessary (it's the
                        #  same?).
                        pipeline_id=self.pipeline_id,
                    ),
                    section_size=self.process.notification_log_section_size,
                )
                # Todo: Support upstream partition IDs different from self.pipeline_id?
                # Todo: Support combining partitions. Read from different partitions
                #  but write to the same partition,
                # could be one os process that reads from many logs of the same
                # upstream app, or many processes each
                # reading one partition with contention writing to the same partition).
                # Todo: Support dividing partitions Read from one but write to many.
                #  Maybe one process per
                # upstream partition, round-robin to pick partition for write. Or
                # have many processes reading
                # with each taking it in turn to skip processing somehow.
                # Todo: Dividing partitions would allow a stream to flow at the same
                #  rate through slower
                # process applications.
                # Todo: Support merging results from "replicated state machines" -
                #  could have a command
                # logging process that takes client commands and presents them in a
                # notification log.
                # Then the system could be deployed in different places, running
                # independently, receiving
                # the same commands, and running the same processes. The command
                # logging process could
                # be accompanied with a result logging process that reads results
                # from replicas as they
                # are available. Not sure what to do if replicas return different
                # things. If one replica
                # goes down, then it could resume by pulling events from another? Not
                # sure what to do.
                # External systems could be modelled as commands.

            # Make the process follow the upstream notification log.
            self.process.follow(upstream_name, notification_log)

        # Subscribe to broadcast prompts published by the process application.
        subscribe(handler=self.broadcast_prompt, predicate=is_prompt_to_pull)

        try:
            self.loop_on_prompts()
        finally:
            unsubscribe(handler=self.broadcast_prompt, predicate=is_prompt_to_pull)

    @retry(CausalDependencyFailed, max_attempts=100, wait=0.1)
    def loop_on_prompts(self) -> None:

        # Run once, in case prompts were missed.
        self.run_process()

        # Loop on getting prompts.
        while True:
            try:
                # Todo: Make the poll interval gradually increase if there are only
                #  timeouts?
                item = self.inbox.get(timeout=self.poll_interval)
                self.inbox.task_done()

                if isinstance(item, PromptToQuit):
                    self.process.close()
                    break

                elif isinstance(item, PromptToPull):
                    self.run_process(item)

                else:
                    raise ProgrammingError("Unsupported prompt: {}".format(item))

            except Empty:
                # Basically, we're polling after a timeout.
                self.run_process()

    @retry((OperationalError, RecordConflictError), max_attempts=100, wait=0.1)
    def run_process(self, prompt: Optional[Prompt] = None) -> None:
        self.process.run(prompt)

    def broadcast_prompt(self, prompt: PromptToPull) -> None:
        if self.outbox is not None:
            self.outbox.put(prompt)



import sys
from unittest import skip, skipIf

from eventsourcing.application.popo import PopoApplication
from eventsourcing.tests.contrib_tests.paxos_tests import test_paxos_system
from eventsourcing.tests.sequenced_item_tests.test_popo_record_manager import (
    PopoTestCase,
)


class TestPaxosSystemWithPopo(PopoTestCase, test_paxos_system.TestPaxosSystem):
    infrastructure_class = PopoApplication

    @skip("Popo doesn't do multiprocessing")
    def test_multiprocessing_performance(self):
        pass

    @skip("Popo doesn't do multiprocessing")
    def test_multiprocessing(self):
        pass



from abc import ABC, abstractmethod
from typing import Any, Generic, Optional, TypeVar


class DatastoreSettings(object):
    """
    Settings for Datastore.
    """


TDatastoreSettings = TypeVar("TDatastoreSettings", bound=DatastoreSettings)


class AbstractDatastore(ABC, Generic[TDatastoreSettings]):
    can_drop_tables = True
    """
    Datastores hold stored event records, used by a record manager.
    """

    def __init__(self, settings: TDatastoreSettings):
        self.settings: TDatastoreSettings = settings

    @property
    def session(self) -> Optional[Any]:
        return None

    @abstractmethod
    def setup_connection(self) -> None:
        """Sets up a connection to a datastore."""

    @abstractmethod
    def close_connection(self) -> None:
        """Drops connection to a datastore."""

    @abstractmethod
    def setup_tables(self) -> None:
        """Sets up tables used to store events."""

    @abstractmethod
    def setup_table(self, table: Any) -> None:
        """Sets up given table."""

    @abstractmethod
    def drop_tables(self) -> None:
        """Drops tables used to store events."""

    @abstractmethod
    def drop_table(self, table: Any) -> None:
        """Drops given table."""

    @abstractmethod
    def truncate_tables(self) -> None:
        """Truncates tables used to store events."""


class DatastoreError(Exception):
    pass


class DatastoreConnectionError(DatastoreError):
    pass


class DatastoreTableError(DatastoreError):
    pass



# import datetime
# import traceback
# import uuid
# from multiprocessing.pool import Pool
# from unittest.case import TestCase, skip
#
# from eventsourcing.contrib.suffixtrees.application import SuffixTreeApplication
# from eventsourcing.contrib.suffixtrees.domain.model.generalizedsuffixtree import GeneralizedSuffixTree, \
#     STRING_ID_END, SuffixTreeEdge, SuffixTreeNode
# from eventsourcing.tests.base import notquick
# from eventsourcing.tests.datastore_tests.test_cassandra import CassandraDatastoreTestCase
# from eventsourcing.tests.unit_test_fixtures_suffix_tree_text import LONG_TEXT, LONG_TEXT_CONT
#
#
# class GeneralizedSuffixTreeTestCase(TestCase):
#     def setUp(self):
#         super(GeneralizedSuffixTreeTestCase, self).setUp()
#         self.app = SuffixTreeApplication(
#             stored_event_repository=PythonObjectsStoredEventRepository()
#         )
#
#     def tearDown(self):
#         self.app.close()
#         super(GeneralizedSuffixTreeTestCase, self).tearDown()
#
#     def add_string_to_suffix_tree(self, string, string_id, suffix_tree):
#         print("Adding string to suffix tree: {}".format(repr(string[:100])))
#         started = datetime.datetime.now()
#         suffix_tree.add_string(string, string_id=string_id)
#         print(" - added string in: {}".format(datetime.datetime.now() - started))
#
#
# @skip
# @notquick
# class TestGeneralizedSuffixTreeFast(GeneralizedSuffixTreeTestCase):
#     def test_empty_string(self):
#         st = self.app.register_new_suffix_tree()
#         assert isinstance(st, GeneralizedSuffixTree)
#         self.add_string_to_suffix_tree('', '1', st)
#         self.assertEqual(self.app.find_substring_edge('not there', st.id), (None, None))
#         self.assertEqual(self.app.find_substring_edge('', st.id), (None, None))
#         self.assertFalse(self.app.has_substring('not there', st.id))
#         self.assertFalse(self.app.has_substring('', st.id))
#
#     def test_add_prefix_of_previous(self):
#         st = self.app.register_new_suffix_tree()
#         assert isinstance(st, GeneralizedSuffixTree)
#         self.add_string_to_suffix_tree('aba', '1', st)
#         self.add_string_to_suffix_tree('a', '2', st)
#
#         # Check the string ID is returned.
#         result_ids = self.app.find_string_ids('a', st.id)
#         self.assertEqual(result_ids, {'1', '2'})
#
#     def test_given_add_duplicate_when_query_with_suffix_then_results_have_both(self):
#         st = self.app.register_new_suffix_tree()
#         assert isinstance(st, GeneralizedSuffixTree)
#         self.add_string_to_suffix_tree('ab', '1', st)
#         self.add_string_to_suffix_tree('ab', '2', st)
#
#         # Check the string ID is returned.
#         result_ids = self.app.find_string_ids('b', st.id)
#         self.assertEqual(result_ids, {'1', '2'})
#
#     def test_extended_copy_of_previous_string(self):
#         st = self.app.register_new_suffix_tree()
#         assert isinstance(st, GeneralizedSuffixTree)
#         self.add_string_to_suffix_tree('a', '1', st)
#         self.add_string_to_suffix_tree('ab', '2', st)
#
#         # Check the common prefix can be found.
#         result_ids = self.app.find_string_ids('a', st.id)
#         self.assertEqual(result_ids, {'1', '2'})
#
#         # Check the string ID is returned.
#         result_ids = self.app.find_string_ids('ab', st.id)
#         self.assertEqual(result_ids, {'2'})
#
#     def test_trucated_copy_of_previous_string(self):
#         st = self.app.register_new_suffix_tree()
#         assert isinstance(st, GeneralizedSuffixTree)
#         self.add_string_to_suffix_tree('1a', '2', st)
#         self.add_string_to_suffix_tree('1ab', '1', st)
#
#         # Check the common prefix can be found.
#         result_ids = self.app.find_string_ids('a', st.id)
#         self.assertEqual(result_ids, {'1', '2'})
#
#         # Check the string ID is returned.
#         result_ids = self.app.find_string_ids('ab', st.id)
#         self.assertEqual(result_ids, {'1'})
#
#         # Check the string ID is returned.
#         result_ids = self.app.find_string_ids('b', st.id)
#         self.assertEqual(result_ids, {'1'})
#
#     def test_extended_copy_of_longer_previous_string(self):
#         st = self.app.register_new_suffix_tree()
#         assert isinstance(st, GeneralizedSuffixTree)
#         self.add_string_to_suffix_tree('ab', '1', st)
#         self.add_string_to_suffix_tree('abc', '2', st)
#
#         # Check the common prefix can be found.
#         result_ids = self.app.find_string_ids('a', st.id)
#         self.assertEqual(result_ids, {'1', '2'})
#
#         # Check the common prefix can be found.
#         result_ids = self.app.find_string_ids('ab', st.id)
#         self.assertEqual(result_ids, {'1', '2'})
#
#         # Check the extention can be found.
#         result_ids = self.app.find_string_ids('c', st.id)
#         self.assertEqual(result_ids, {'2'})
#
#         # Check the extended string can be found.
#         result_ids = self.app.find_string_ids('abc', st.id)
#         self.assertEqual(result_ids, {'2'})
#
#     def test_non_repeating_string(self):
#         st = self.app.register_new_suffix_tree()
#         assert isinstance(st, GeneralizedSuffixTree)
#         string_id = '1'
#         string = 'abc'
#         self.add_string_to_suffix_tree(string, string_id, st)
#
#         # Check various non-substrings aren't in the tree.
#         self.assertEqual(self.app.find_substring_edge('not there', st.id), (None, None))
#         self.assertEqual(self.app.find_substring_edge('', st.id), (None, None))
#         self.assertFalse(self.app.has_substring('not there', st.id))
#         self.assertFalse(self.app.has_substring('', st.id))
#
#         # Check substrings are in the tree.
#         edge, ln = self.app.find_substring_edge('abc', st.id)
#         assert isinstance(edge, SuffixTreeEdge)
#         # self.assertEqual(edge.label, string + STRING_ID_END)
#         # self.assertEqual(edge.label, string + string_id + STRING_ID_END)
#
#         # Check the edge has a node.
#         node = self.app.node_repo[edge.dest_node_id]
#         assert isinstance(node, SuffixTreeNode)
#
#         # Check the node has a string ID.
#         self.assertEqual(node.string_id, string_id)
#
#         # Check the string ID is returned.
#         result_ids = self.app.find_string_ids(string, st.id)
#         self.assertEqual(result_ids, {string_id})
#
#     def test_repeating_string(self):
#         st = self.app.register_new_suffix_tree()
#         self.add_string_to_suffix_tree("aaa", '1', st)
#
#         edge, ln = self.app.find_substring_edge('a', st.id)
#         # self.assertEqual(edge.label, 'a')
#         edge, ln = self.app.find_substring_edge('aa', st.id)
#         # self.assertEqual(edge.label, 'a')
#         edge, ln = self.app.find_substring_edge('aaa', st.id)
#         # self.assertEqual(edge.label, 'a1' + STRING_ID_END)
#         edge, ln = self.app.find_substring_edge('b', st.id)
#         self.assertEqual(edge, None)
#         self.assertTrue(self.app.has_substring('a', st.id))
#         self.assertTrue(self.app.has_substring('aa', st.id))
#         self.assertTrue(self.app.has_substring('aaa', st.id))
#         self.assertFalse(self.app.has_substring('aaaa', st.id))
#         self.assertFalse(self.app.has_substring('b', st.id))
#         # case sensitive by default
#         self.assertFalse(self.app.has_substring('A', st.id))
#
#         # Check ID is returned.
#         self.assertEqual(self.app.find_string_ids('a', st.id), {'1'})
#         self.assertEqual(self.app.find_string_ids('aa', st.id), {'1'})
#         self.assertEqual(self.app.find_string_ids('aaa', st.id), {'1'})
#
#     def test_two_identical_strings(self):
#         st = self.app.register_new_suffix_tree()
#         assert isinstance(st, GeneralizedSuffixTree)
#         string1_id = '1'
#         string1 = 'abc'
#         string2_id = '2'
#         string2 = 'abc'
#         self.add_string_to_suffix_tree(string1, string1_id, st)
#         self.add_string_to_suffix_tree(string2, string2_id, st)
#
#         # Check various non-substrings aren't in the tree.
#         self.assertEqual(self.app.find_substring_edge('not there', st.id), (None, None))
#         self.assertEqual(self.app.find_substring_edge('', st.id), (None, None))
#         self.assertFalse(self.app.has_substring('not there', st.id))
#         self.assertFalse(self.app.has_substring('', st.id))
#
#         # Check substrings of string1 are in the tree.
#         edge, ln = self.app.find_substring_edge(string1, st.id)
#         assert isinstance(edge, SuffixTreeEdge)
#         # self.assertEqual(edge.label, string1 + STRING_ID_END)
#         # self.assertEqual(edge.label, string1 + string1_id + STRING_ID_END)
#         # self.assertEqual(edge.label, string1 + string1_id )
#
#         # Check the edge has a node.
#         node = self.app.node_repo[edge.dest_node_id]
#         assert isinstance(node, SuffixTreeNode)
#
#         # Check the node doesn't have a string ID.
#         # self.assertEqual(node.string_id, None)
#
#         # Check both IDs are returned.
#         result_ids = self.app.find_string_ids(string1, st.id)
#         self.assertEqual(result_ids, {string1_id, string2_id}, (result_ids))
#
#     def test_two_non_repeating_strings(self):
#         st = self.app.register_new_suffix_tree()
#         assert isinstance(st, GeneralizedSuffixTree)
#         string1_id = '1'
#         string1 = 'abc'
#         string2_id = '2'
#         string2 = 'def'
#         self.add_string_to_suffix_tree(string1, string1_id, st)
#         self.add_string_to_suffix_tree(string2, string2_id, st)
#
#         # Check various non-substrings aren't in the tree.
#         self.assertEqual(self.app.find_substring_edge('not there', st.id), (None, None))
#         self.assertEqual(self.app.find_substring_edge('', st.id), (None, None))
#         self.assertFalse(self.app.has_substring('not there', st.id))
#         self.assertFalse(self.app.has_substring('', st.id))
#
#         # Check substrings of string1 are in the tree.
#         edge, ln = self.app.find_substring_edge(string1, st.id)
#         assert isinstance(edge, SuffixTreeEdge)
#         # self.assertEqual(edge.label, string1 + STRING_ID_END)
#         # self.assertEqual(edge.label, string1 + string1_id + STRING_ID_END)
#
#         # Check the edge has a node.
#         node = self.app.node_repo[edge.dest_node_id]
#         assert isinstance(node, SuffixTreeNode)
#
#         # Check the node has a string ID.
#         self.assertEqual(node.string_id, string1_id)
#
#         # Check the string1 ID is returned.
#         result_ids = self.app.find_string_ids(string1, st.id)
#         self.assertEqual(result_ids, {string1_id})
#
#         # Check substrings of string2 are in the tree.
#         edge, ln = self.app.find_substring_edge(string2, st.id)
#         assert isinstance(edge, SuffixTreeEdge)
#         # self.assertEqual(edge.label, string2 + STRING_ID_END)
#         # self.assertEqual(edge.label, string2 + string2_id+ STRING_ID_END)
#
#         # Check the edge has a node.
#         node = self.app.node_repo[edge.dest_node_id]
#         assert isinstance(node, SuffixTreeNode)
#
#         # Check the node has a string ID.
#         self.assertEqual(node.string_id, string2_id)
#
#         # Check the string2 ID is returned.
#         result_ids = self.app.find_string_ids(string2, st.id)
#         self.assertEqual(result_ids, {string2_id})
#
#     def test_common_prefix_two_non_repeating_strings_with(self):
#         st = self.app.register_new_suffix_tree()
#         assert isinstance(st, GeneralizedSuffixTree)
#         common_prefix = 'z'
#         string1_id = '1'
#         string1 = common_prefix + 'abc'
#         string2_id = '2'
#         string2 = common_prefix + 'def'
#         self.add_string_to_suffix_tree(string1, string1_id, st)
#         self.add_string_to_suffix_tree(string2, string2_id, st)
#
#         # Check various non-substrings aren't in the tree.
#         self.assertEqual(self.app.find_substring_edge('not there', st.id), (None, None))
#         self.assertEqual(self.app.find_substring_edge('', st.id), (None, None))
#         self.assertFalse(self.app.has_substring('not there', st.id))
#         self.assertFalse(self.app.has_substring('', st.id))
#
#         # Check substrings of string1 are in the tree.
#         edge, ln = self.app.find_substring_edge(string1, st.id)
#         assert isinstance(edge, SuffixTreeEdge)
#         # self.assertEqual(edge.label, string1[1:] + STRING_ID_END)
#         # self.assertEqual(edge.label, string1[1:] + string1_id + STRING_ID_END)
#
#         # Check the edge has a node.
#         node = self.app.node_repo[edge.dest_node_id]
#         assert isinstance(node, SuffixTreeNode)
#
#         # Check the node has a string ID.
#         self.assertEqual(node.string_id, string1_id)
#
#         # Check the string1 ID is returned.
#         result_ids = self.app.find_string_ids(string1, st.id)
#         self.assertEqual(result_ids, {string1_id})
#
#         # Check substrings of string2 are in the tree.
#         edge, ln = self.app.find_substring_edge(string2, st.id)
#         assert isinstance(edge, SuffixTreeEdge)
#         # self.assertEqual(edge.label, string2[1:] + STRING_ID_END)
#         # self.assertEqual(edge.label, string2[1:] + string2_id + STRING_ID_END)
#
#         # Check the edge has a node.
#         node = self.app.node_repo[edge.dest_node_id]
#         assert isinstance(node, SuffixTreeNode)
#
#         # Check the node has a string ID.
#         self.assertEqual(node.string_id, string2_id)
#
#         # Check the string2 ID is returned.
#         result_ids = self.app.find_string_ids(string2, st.id)
#         self.assertEqual(result_ids, {string2_id})
#
#         # Check common prefix is in the tree.
#         edge, ln = self.app.find_substring_edge(common_prefix, st.id)
#         assert isinstance(edge, SuffixTreeEdge)
#         self.assertEqual(edge.label, common_prefix)
#
#         # Check the edge has a node.
#         node = self.app.node_repo[edge.dest_node_id]
#         assert isinstance(node, SuffixTreeNode)
#
#         # Check the node has a string ID.
#         self.assertEqual(node.string_id, None)
#
#         # Check both IDs are returned.
#         result_ids = self.app.find_string_ids(common_prefix, st.id)
#         self.assertEqual(result_ids, {string1_id, string2_id})
#
#     def test_common_suffix_two_non_repeating_strings_with(self):
#         st = self.app.register_new_suffix_tree()
#         assert isinstance(st, GeneralizedSuffixTree)
#         common_suffix = 'z'
#         string1_id = '1'
#         string1 = 'abc' + common_suffix
#         string2_id = '2'
#         string2 = 'def' + common_suffix
#         self.add_string_to_suffix_tree(string1, string1_id, st)
#         self.add_string_to_suffix_tree(string2, string2_id, st)
#
#         # Check various non-substrings aren't in the tree.
#         self.assertEqual(self.app.find_substring_edge('not there', st.id), (None, None))
#         self.assertEqual(self.app.find_substring_edge('', st.id), (None, None))
#         self.assertFalse(self.app.has_substring('not there', st.id))
#         self.assertFalse(self.app.has_substring('', st.id))
#
#         # Check substrings of string1 are in the tree.
#         edge, ln = self.app.find_substring_edge(string1, st.id)
#         assert isinstance(edge, SuffixTreeEdge)
#         # self.assertEqual(edge.label, string1 + STRING_ID_END)
#         # self.assertEqual(edge.label, string1 + string1_id + STRING_ID_END)
#
#         # Check the edge has a node.
#         node = self.app.node_repo[edge.dest_node_id]
#         assert isinstance(node, SuffixTreeNode)
#
#         # Check the node has a string ID.
#         self.assertEqual(node.string_id, string1_id)
#
#         # Check the string1 ID is returned.
#         result_ids = self.app.find_string_ids(string1, st.id)
#         self.assertEqual(result_ids, {string1_id})
#
#         # Check substrings of string2 are in the tree.
#         edge, ln = self.app.find_substring_edge(string2, st.id)
#         assert isinstance(edge, SuffixTreeEdge)
#         # self.assertEqual(edge.label, string2 + STRING_ID_END)
#         # self.assertEqual(edge.label, string2 + string2_id + STRING_ID_END)
#
#         # Check the edge has a node.
#         node = self.app.node_repo[edge.dest_node_id]
#         assert isinstance(node, SuffixTreeNode)
#
#         # Check the node has a string ID.
#         self.assertEqual(node.string_id, string2_id)
#
#         # Check the string2 ID is returned.
#         result_ids = self.app.find_string_ids(string2, st.id)
#         self.assertEqual(result_ids, {string2_id})
#
#         # Check common suffix is in the tree.
#         edge, ln = self.app.find_substring_edge(common_suffix, st.id)
#         assert isinstance(edge, SuffixTreeEdge)
#         # self.assertEqual(edge.label, common_suffix + STRING_ID_END)
#         # self.assertEqual(edge.label, common_suffix)
#
#         # Check the edge has a node.
#         node = self.app.node_repo[edge.dest_node_id]
#         assert isinstance(node, SuffixTreeNode)
#
#         # Check the node doesn't have a string ID.
#         # self.assertEqual(node.string_id, None)
#
#         # Check both IDs are returned.
#         result_ids = self.app.find_string_ids(common_suffix, st.id)
#         self.assertEqual(result_ids, {string1_id, string2_id})
#
#     def test_mississippi(self):
#         st = self.app.register_new_suffix_tree()
#         self.add_string_to_suffix_tree("mississippi", "$", st)
#         self.add_string_to_suffix_tree("mudpie", "#", st)
#         self.add_string_to_suffix_tree("ball", "%", st)
#         self.add_string_to_suffix_tree("ball", "&", st)
#
#         # Check there isn't an 'a'.
#         self.assertEqual(self.app.find_substring_edge('j', st.id), (None, None))
#
#         # Check 'm'.
#         edge, ln = self.app.find_substring_edge('m', st.id)
#         # self.assertEqual(edge.label, 'm')
#         self.assertEqual(self.app.find_string_ids('m', st.id), {'$', '#'})
#
#         # Check 'mi'.
#         edge, ln = self.app.find_substring_edge('mi', st.id)
#         # self.assertEqual(edge.label, 'ississippi' + STRING_ID_END)
#         self.assertEqual(self.app.find_string_ids('mi', st.id), {'$'})
#
#         # Check 'mu'.
#         edge, ln = self.app.find_substring_edge('mu', st.id)
#         # self.assertEqual(edge.label, 'udpie' + STRING_ID_END)
#         self.assertEqual(self.app.find_string_ids('mu', st.id), {'#'})
#
#         # Check 'missi'.
#         edge, ln = self.app.find_substring_edge('missi', st.id)
#         self.assertEqual(edge.label, 'ississippi' + STRING_ID_END)
#
#         # Check 'issi'.
#         edge, ln = self.app.find_substring_edge('issi', st.id)
#         self.assertEqual(edge.label, 'ssi')
#
#         # Check 'is'.
#         edge, ln = self.app.find_substring_edge('is', st.id)
#         self.assertEqual(edge.label, 'ssi')
#
#         # Check 'si'.
#         edge, ln = self.app.find_substring_edge('si', st.id)
#         self.assertEqual(edge.label, 'i')
#
#         # Check 'issip'.
#         edge, ln = self.app.find_substring_edge('issip', st.id)
#         self.assertEqual(edge.label, 'ppi' + STRING_ID_END)
#
#         # Check 'ssip'.
#         edge, ln = self.app.find_substring_edge('ssip', st.id)
#         self.assertEqual(edge.label, 'ppi' + STRING_ID_END)
#
#         # Check 'sip'.
#         edge, ln = self.app.find_substring_edge('sip', st.id)
#         self.assertEqual(edge.label, 'ppi' + STRING_ID_END)
#
#         # Check 'ip'.
#         edge, ln = self.app.find_substring_edge('ip', st.id)
#         self.assertEqual(edge.label, 'ppi' + STRING_ID_END)
#
#         # Check 'i'.
#         edge, ln = self.app.find_substring_edge('i', st.id)
#         self.assertEqual(edge.label, 'i')
#
#         # Check 'ippi'.
#         edge, ln = self.app.find_substring_edge('ippi', st.id)
#         self.assertEqual(edge.label, 'ppi' + STRING_ID_END)
#
#         # Check 'mudpie'.
#         edge, ln = self.app.find_substring_edge('mudpie', st.id)
#         self.assertEqual(edge.label, 'udpie' + STRING_ID_END)
#
#         # Check 'ball'.
#         edge, ln = self.app.find_substring_edge('ball', st.id)
#         self.assertEqual(edge.label, 'ball' + STRING_ID_END)
#
#         # Check ID is returned.
#         self.assertEqual(self.app.find_string_ids('mi', st.id), {'$'})
#         self.assertEqual(self.app.find_string_ids('si', st.id), {'$'})
#         self.assertEqual(self.app.find_string_ids('pp', st.id), {'$'})
#         self.assertEqual(self.app.find_string_ids('mu', st.id), {'#'})
#         self.assertEqual(self.app.find_string_ids('m', st.id), {'$', '#'})
#         self.assertEqual(self.app.find_string_ids('b', st.id), {'%', '&'})
#
#     def test_colours(self):
#         st = self.app.register_new_suffix_tree()
#         self.add_string_to_suffix_tree("blue", "$", st)
#         self.add_string_to_suffix_tree("red", "#", st)
#
#         # Check 'b'.
#         edge, ln = self.app.find_substring_edge('b', st.id)
#         self.assertEqual(edge.label, 'blue' + STRING_ID_END)
#
#         # Check 'l'.
#         edge, ln = self.app.find_substring_edge('l', st.id)
#         self.assertEqual(edge.label, 'lue' + STRING_ID_END)
#
#         # Check 'u'.
#         edge, ln = self.app.find_substring_edge('u', st.id)
#         self.assertEqual(edge.label, 'ue' + STRING_ID_END)
#
#         # Check 'e'.
#         edge, ln = self.app.find_substring_edge('e', st.id)
#         self.assertEqual(edge.label, 'e')
#
#         # Check 'ue'.
#         edge, ln = self.app.find_substring_edge('ue', st.id)
#         self.assertEqual(edge.label, 'ue' + STRING_ID_END)
#
#         # Check 'lue'.
#         edge, ln = self.app.find_substring_edge('lue', st.id)
#         self.assertEqual(edge.label, 'lue' + STRING_ID_END)
#
#         # Check 'blue'.
#         edge, ln = self.app.find_substring_edge('blue', st.id)
#         self.assertEqual(edge.label, 'blue' + STRING_ID_END)
#
#         # Check 're'.
#         edge, ln = self.app.find_substring_edge('re', st.id)
#         self.assertEqual(edge.label, 'red' + STRING_ID_END)
#
#         # Check 'ed'.
#         edge, ln = self.app.find_substring_edge('ed', st.id)
#         self.assertEqual(edge.label, 'd' + STRING_ID_END)
#
#         # Check 'red'.
#         edge, ln = self.app.find_substring_edge('red', st.id)
#         self.assertEqual(edge.label, 'red' + STRING_ID_END)
#
#     def test_find_string_ids(self):
#         # This test is the first to involve the children of nodes.
#         st = self.app.register_new_suffix_tree()
#         string1_id = uuid.uuid4().hex
#         string2_id = uuid.uuid4().hex
#         string3_id = uuid.uuid4().hex
#         # string4_id = uuid.uuid4().hex
#         self.add_string_to_suffix_tree("blue", string1_id, st)
#         self.add_string_to_suffix_tree("red", string2_id, st)
#         self.add_string_to_suffix_tree("blues", string3_id, st)
#
#         strings = self.app.find_string_ids('$', st.id)
#         assert not strings, strings
#         strings = self.app.find_string_ids('0', st.id)
#         assert not strings, strings
#         strings = self.app.find_string_ids('1', st.id)
#         assert not strings, strings
#         strings = self.app.find_string_ids('2', st.id)
#         assert not strings, strings
#         strings = self.app.find_string_ids('3', st.id)
#         assert not strings, strings
#         strings = self.app.find_string_ids('4', st.id)
#         assert not strings, strings
#         strings = self.app.find_string_ids('5', st.id)
#         assert not strings, strings
#         strings = self.app.find_string_ids('6', st.id)
#         assert not strings, strings
#         strings = self.app.find_string_ids('7', st.id)
#         assert not strings, strings
#
#         # Find 'b'.
#         strings = self.app.find_string_ids('b', st.id)
#         self.assertIn(string1_id, strings)
#         self.assertNotIn(string2_id, strings)
#
#         # Find 'e'.
#         strings = self.app.find_string_ids('e', st.id)
#         self.assertEqual(len(strings), 3)
#         self.assertIn(string1_id, strings, (string1_id, string2_id, string3_id, strings))
#         self.assertIn(string2_id, strings, (string1_id, string2_id, string3_id, strings))
#         self.assertIn(string3_id, strings, (string1_id, string2_id, string3_id, strings))
#
#         # Find 'e' - limit 1.
#         strings = self.app.find_string_ids('e', st.id, limit=1)
#         self.assertEqual(len(strings), 1)
#
#         # Find 'e' - limit 2.
#         strings = self.app.find_string_ids('e', st.id, limit=2)
#         self.assertEqual(len(strings), 2)
#
#         # Find 'r'.
#         strings = self.app.find_string_ids('r', st.id)
#         self.assertNotIn(string1_id, strings)
#         self.assertIn(string2_id, strings)
#
#         # Find 'd'.
#         strings = self.app.find_string_ids('d', st.id)
#         self.assertNotIn(string1_id, strings)
#         self.assertIn(string2_id, strings)
#
#         # Find 's'.
#         strings = self.app.find_string_ids('s', st.id)
#         self.assertNotIn(string1_id, strings)
#         self.assertNotIn(string2_id, strings)
#         self.assertIn(string3_id, strings)
#
#     def test_add_string_to_suffixtree_from_repo(self):
#         # Check adding strings after getting the tree from the repo.
#         st = self.app.register_new_suffix_tree()
#
#         st = self.app.get_suffix_tree(st.id)
#         self.add_string_to_suffix_tree('blue', '1', st)
#
#         st = self.app.get_suffix_tree(st.id)
#         self.add_string_to_suffix_tree('green', '2', st)
#
#         st = self.app.get_suffix_tree(st.id)
#         self.add_string_to_suffix_tree('yellow', '3', st)
#
#         st = self.app.get_suffix_tree(st.id)
#         strings_ids = self.app.find_string_ids('e', st.id)
#         self.assertEqual(3, len(strings_ids), strings_ids)
#         self.assertEqual(['1', '2', '3'], sorted(strings_ids))
#
#     def test_remove_string_from_suffixtree(self):
#         st = self.app.register_new_suffix_tree()
#
#         # Add 'blue' and 'green'.
#         self.add_string_to_suffix_tree('blue', '1', st)
#         self.add_string_to_suffix_tree('green', '2', st)
#
#         strings_ids = self.app.find_string_ids('b', st.id)
#         self.assertEqual(1, len(strings_ids))
#
#         strings_ids = self.app.find_string_ids('l', st.id)
#         self.assertEqual(1, len(strings_ids))
#
#         strings_ids = self.app.find_string_ids('u', st.id)
#         self.assertEqual(1, len(strings_ids))
#
#         strings_ids = self.app.find_string_ids('r', st.id)
#         self.assertEqual(1, len(strings_ids))
#
#         strings_ids = self.app.find_string_ids('g', st.id)
#         self.assertEqual(1, len(strings_ids))
#
#         strings_ids = self.app.find_string_ids('e', st.id)
#         self.assertEqual(2, len(strings_ids))
#
#         # Remove 'blue'.
#         st.remove_string('blue', '1')
#
#         strings_ids = self.app.find_string_ids('b', st.id)
#         self.assertEqual(0, len(strings_ids))
#
#         strings_ids = self.app.find_string_ids('l', st.id)
#         self.assertEqual(0, len(strings_ids))
#
#         strings_ids = self.app.find_string_ids('u', st.id)
#         self.assertEqual(0, len(strings_ids))
#
#         strings_ids = self.app.find_string_ids('r', st.id)
#         self.assertEqual(1, len(strings_ids))
#
#         strings_ids = self.app.find_string_ids('g', st.id)
#         self.assertEqual(1, len(strings_ids))
#
#         strings_ids = self.app.find_string_ids('e', st.id)
#         self.assertEqual(1, len(strings_ids), strings_ids)
#
#         # Add 'blue' again.
#         self.add_string_to_suffix_tree('blue', '1', st)
#
#         strings_ids = self.app.find_string_ids('b', st.id)
#         self.assertEqual(1, len(strings_ids))
#
#         strings_ids = self.app.find_string_ids('l', st.id)
#         self.assertEqual(1, len(strings_ids))
#
#         strings_ids = self.app.find_string_ids('u', st.id)
#         self.assertEqual(1, len(strings_ids))
#
#         strings_ids = self.app.find_string_ids('r', st.id)
#         self.assertEqual(1, len(strings_ids))
#
#         strings_ids = self.app.find_string_ids('g', st.id)
#         self.assertEqual(1, len(strings_ids))
#
#         strings_ids = self.app.find_string_ids('e', st.id)
#         self.assertEqual(2, len(strings_ids))
#
#
# @skip
# @notquick
# class TestGeneralizedSuffixTreeSlow(GeneralizedSuffixTreeTestCase):
#     def test_long_string(self):
#         st = self.app.register_new_suffix_tree()
#         self.add_string_to_suffix_tree(LONG_TEXT[:12000], '1', st)
#         self.assertEqual(self.app.find_string_ids('Ukkonen', st.id), {'1'})
#         self.assertEqual(self.app.find_string_ids('Optimal', st.id), {'1'})
#         self.assertFalse(self.app.find_string_ids('ukkonen', st.id))
#         self.add_string_to_suffix_tree(LONG_TEXT_CONT[:1000], '2', st)
#         self.assertEqual(self.app.find_string_ids('Burrows-Wheeler', st.id), {'2'})
#         self.assertEqual(self.app.find_string_ids('suffix', st.id), {'1', '2'})
#
#     def test_split_long_string(self):
#         # Split the long string into separate strings, and make some IDs.
#         list_of_strings = [w for w in LONG_TEXT[:1000].split(' ') if w]
#
#         print("Long string split into words: {}".format(list_of_strings))
#
#         string_ids = {}
#         for string in list_of_strings:
#             string_id = uuid.uuid4().hex
#             string_ids[string_id] = string
#
#         # Build the suffix tree.
#         st = self.app.register_new_suffix_tree()
#         for string_id, string in string_ids.items():
#             self.add_string_to_suffix_tree(string, string_id, st)
#
#         # Check the suffix tree.
#         for string_id, string in string_ids.items():
#             self.assertIn(string_id, self.app.find_string_ids(string, st.id), (string, string_id))
#
#         # Build the suffix tree.
#         #  - this time, get the suffix tree afresh from the repo each time
#         st = self.app.register_new_suffix_tree()
#         for string_id, string in string_ids.items():
#             st = self.app.get_suffix_tree(st.id)
#             self.add_string_to_suffix_tree(string, string_id, st)
#
#         # Check the suffix tree.
#         for string_id, string in string_ids.items():
#             self.assertIn(string_id, self.app.find_string_ids(string, st.id))
#
#     def test_case_insensitivity(self):
#         st = self.app.register_new_suffix_tree(case_insensitive=True)
#         self.add_string_to_suffix_tree(LONG_TEXT[:12000], '1', st)
#         self.assertEqual(self.app.find_string_ids('ukkonen', st.id), {'1'})
#         self.assertEqual(self.app.find_string_ids('Optimal', st.id), {'1'})
#         self.assertEqual(self.app.find_string_ids('burrows-wheeler', st.id), set())
#         self.add_string_to_suffix_tree(LONG_TEXT_CONT[:1000], '2', st)
#         self.assertEqual(self.app.find_string_ids('ukkonen', st.id), {'1'})
#         self.assertEqual(self.app.find_string_ids('Optimal', st.id), {'1'})
#         self.assertEqual(self.app.find_string_ids('burrows-wheeler', st.id), {'2'})
#
#
# @skip
# @notquick
# class TestMultiprocessingWithGeneralizedSuffixTree(CassandraDatastoreTestCase):
#     def setUp(self):
#         super(TestMultiprocessingWithGeneralizedSuffixTree, self).setUp()
#         self.app = None
#
#     def tearDown(self):
#         super(TestMultiprocessingWithGeneralizedSuffixTree, self).tearDown()
#         if self.app is not None:
#             self.app.close()
#
#     @notquick
#     def test_words_in_sorted_order(self):
#         self.check_words(is_sorted=True)
#
#     # Todo: Fix this - adding strings in a random order sometimes breaks (perhaps a dict is causing indeterminate
#     # order).
#     # def test_words_in_unsorted_order(self):
#     #     self.check_words()
#
#     # Todo: Fix this - adding strings in a reversed sorted order always fails. Not sure why all substrings of 'ree'
#     # fail. The suffix is obviously not moving along in the same way as it does when the nodes are added. Perhaps it
#     #  needs to add the IDs when explicit match is made, and then move the first char along by one? Not sure so
#     # trace it out?
#     # def test_words_in_reverse_sorted_order(self):
#     #     self.check_words(is_reversed=True)
#     #
#     # The error reported is:-
#     #
#     # >       self.assertFalse(errors, "\n".join(errors))
#     # E       Not found: substring ''e'' from string ''tree''
#     # E       Not found: substring ''ee'' from string ''tree''
#     # E       Not found: substring ''r'' from string ''tree''
#     # E       Not found: substring ''re'' from string ''tree''
#     # E       Not found: substring ''ree'' from string ''tree''
#
#     def check_words(self, is_sorted=False, is_reversed=False):
#         # Split the long string into separate strings, and make some IDs.
#         words = list([w for w in LONG_TEXT[:100].split(' ') if w])
#
#         print("Adding words: {}".format(words))
#
#         # Avoid adding the same string twice (or a prefix of a previous string).
#         #  - because it's a current problem unless we append string IDs, which makes things too slow
#         # words = set(words)
#         # words = [w for w in words if 0 != sum([x.startswith(w) for x in words if x != w])]
#
#         assert words
#
#         # Make a string ID for each string.
#         strings = {}
#         for string in words:
#             string_id = uuid.uuid4().hex
#             strings[string_id] = string
#
#         # Create a new suffix tree.
#         self.app = SuffixTreeApplicationWithCassandra()
#         st = self.app.register_new_suffix_tree()
#         assert st.id in self.app.suffix_tree_repo
#
#         # Close the app, so the pool doesn't inherit it.
#         self.app.close()
#
#         # Start the pool.
#         pool = Pool(initializer=pool_initializer, processes=1)
#
#         words = [[s, sid, st.id] for sid, s in strings.items() if s]
#
#         if is_sorted:
#             words = sorted(words)
#         if is_reversed:
#             words = reversed(words)
#
#         results = pool.map(add_string_to_suffix_tree, words)
#         for result in results:
#             if isinstance(result, Exception):
#                 print(result.args[0][1])
#                 raise result
#
#         # Creat the app again.
#         self.app = SuffixTreeApplicationWithCassandra()
#
#         errors = []
#
#         # Check the suffix tree returns string ID for all substrings of string.
#         for string_id, string in strings.items():
#             # Check all prefixes and suffixes.
#             substrings = sorted(list(get_all_substrings(string)))
#             print("")
#             print("Checking for all substrings of string '{}': {}".format(repr(string),
#                                                                           " ".join([repr(s) for s in substrings])))
#             for substring in substrings:
#                 results = self.app.find_string_ids(substring, st.id)
#                 if string_id not in results:
#                     msg = "Not found: substring '{}' from string '{}'".format(repr(substring), repr(string))
#                     print(msg)
#                     errors.append(msg)
#
#         # Check for errors.
#         self.assertFalse(errors, "\n".join(errors))
#
#
# def get_all_substrings(s):
#     length = len(s)
#     return (s[i:j] for i in range(length) for j in range(i + 1, length + 1))
#
#
# worker_app = None
#
#
# def pool_initializer():
#     global worker_app
#     worker_app = SuffixTreeApplicationWithCassandra()
#
#
# def add_string_to_suffix_tree(args):
#     # random.seed()
#     string, string_id, suffix_tree_id = args
#     print("")
#     print("Adding string to suffix tree: {}: {}".format(string_id, repr(string[:100])))
#     try:
#         assert isinstance(worker_app, AbstractSuffixTreeApplication)
#         suffix_tree = worker_app.get_suffix_tree(suffix_tree_id)
#         assert isinstance(suffix_tree, GeneralizedSuffixTree)
#         started = datetime.datetime.now()
#         suffix_tree.add_string(string, string_id)
#         print(" - added string in: {}".format(datetime.datetime.now() - started))
#     except Exception as e:
#         msg = traceback.format_exc()
#         print(" - failed to add string: {}".format(msg))
#         return Exception((e, msg, string, string_id))
#     return string_id



import json
import uuid
from abc import abstractmethod
from time import sleep
from uuid import uuid4

from eventsourcing.application.policies import PersistencePolicy
from eventsourcing.domain.model.entity import VersionedEntity
from eventsourcing.domain.model.events import (
    EventWithOriginatorID,
    EventWithOriginatorVersion,
    EventWithTimestamp,
    LoggedEvent,
)
from eventsourcing.domain.model.snapshot import Snapshot
from eventsourcing.exceptions import OperationalError, RecordConflictError
from eventsourcing.infrastructure.base import BaseRecordManager
from eventsourcing.infrastructure.eventstore import EventStore
from eventsourcing.infrastructure.iterators import (
    SequencedItemIterator,
    ThreadedSequencedItemIterator,
)
from eventsourcing.infrastructure.sequenceditem import SequencedItem
from eventsourcing.infrastructure.sequenceditemmapper import SequencedItemMapper
from eventsourcing.tests.datastore_tests.base import AbstractDatastoreTestCase
from eventsourcing.utils.times import decimaltimestamp
from eventsourcing.utils.topic import get_topic


class RecordManagerTestCase(AbstractDatastoreTestCase):
    """
    Abstract test case for record managers.
    """

    def __init__(self, *args, **kwargs):
        super(RecordManagerTestCase, self).__init__(*args, **kwargs)
        self._record_manager = None

    def setUp(self):
        super(RecordManagerTestCase, self).setUp()
        if self.datastore is not None:
            self.datastore.setup_connection()
            try:
                self.datastore.setup_tables()
            except:
                self.datastore.drop_tables()
                self.datastore.setup_tables()

    def tearDown(self):
        self._record_manager = None
        if self.datastore is not None:
            self.datastore.drop_tables()
            self.datastore.close_connection()
        super(RecordManagerTestCase, self).tearDown()

    def test(self):
        sequence_id1 = uuid.uuid1()
        sequence_id2 = uuid.uuid1()

        if self.record_manager.contiguous_record_ids:
            old_notifications = list(self.record_manager.get_notification_records())
        else:
            old_notifications = []

        # Check record manager returns empty list when there aren't any items.
        self.assertEqual(self.record_manager.list_items(sequence_id1), [])

        # Check record manager returns no sequence IDs.
        self.assertEqual([], self.record_manager.list_sequence_ids())

        position1, position2, position3 = self.construct_positions()

        self.assertLess(position1, position2)
        self.assertLess(position2, position3)

        # Append an item.
        state1 = json.dumps({"name": "value1"}).encode("utf-8")
        item1 = SequencedItem(
            sequence_id=sequence_id1,
            position=position1,
            topic=self.EXAMPLE_EVENT_TOPIC1,
            state=state1,
        )
        self.record_manager.record_item(item1)

        if self.record_manager.can_list_sequence_ids:
            # Check record manager returns one sequence ID.
            self.assertEqual([sequence_id1], self.record_manager.list_sequence_ids())

        # Append an item to a different sequence.
        state2 = json.dumps({"name": "value2"}).encode("utf-8")
        item2 = SequencedItem(
            sequence_id=sequence_id2,
            position=position1,
            topic=self.EXAMPLE_EVENT_TOPIC1,
            state=state2,
        )
        self.record_manager.record_item(item2)

        # Check record manager returns two sequence IDs.
        if self.record_manager.can_list_sequence_ids:
            self.assertEqual(
                sorted([sequence_id1, sequence_id2]),
                sorted(self.record_manager.list_sequence_ids()),
            )

        # Get first event.
        retrieved_item = self.record_manager.get_item(sequence_id1, position1)
        self.assertEqual(sequence_id1, retrieved_item.sequence_id)
        self.assertEqual(state1, retrieved_item.state)

        # Check index error is raised when item does not exist at position.
        with self.assertRaises(IndexError):
            self.record_manager.get_item(sequence_id1, position2)

        # Check record manager returns the item.
        retrieved_items = self.record_manager.list_items(sequence_id1)
        self.assertEqual(1, len(retrieved_items), str(retrieved_items))
        self.assertIsInstance(retrieved_items[0], SequencedItem)
        self.assertEqual(retrieved_items[0].sequence_id, item1.sequence_id)
        self.assertEqual(retrieved_items[0].state, item1.state)
        self.assertEqual(retrieved_items[0].topic, item1.topic)

        # Check raises RecordConflictError when appending an item at same position in same sequence.
        state3 = json.dumps({"name": "value3"}).encode("utf-8")
        item3 = SequencedItem(
            sequence_id=item1.sequence_id,
            position=position1,
            topic=self.EXAMPLE_EVENT_TOPIC2,
            state=state3,
        )
        self.assertEqual(item1.sequence_id, item3.sequence_id)
        self.assertNotEqual(item1.topic, item3.topic)
        self.assertNotEqual(item1.state, item3.state)
        # - append conflicting item as single item
        with self.assertRaises(RecordConflictError):
            self.record_manager.record_item(item3)

        item4 = SequencedItem(
            sequence_id=item1.sequence_id,
            position=position2,
            topic=self.EXAMPLE_EVENT_TOPIC2,
            state=state3,
        )
        item5 = SequencedItem(
            sequence_id=item1.sequence_id,
            position=position3,
            topic=self.EXAMPLE_EVENT_TOPIC2,
            state=state3,
        )
        # - append conflicting item in list with new items (none should be appended)
        with self.assertRaises(RecordConflictError):
            self.record_manager.record_items([item3, item4, item5])

        # Check there is still only one item.
        retrieved_items = self.record_manager.list_items(sequence_id1)
        self.assertEqual(len(retrieved_items), 1)

        # Check adding an empty list does nothing.
        self.record_manager.record_items([])
        retrieved_items = self.record_manager.list_items(sequence_id1)
        self.assertEqual(len(retrieved_items), 1)

        # Append a second and third item at the next positions.
        self.record_manager.record_items([item4, item5])

        # Check there are three items.
        retrieved_items = self.record_manager.list_items(sequence_id1)
        self.assertEqual(len(retrieved_items), 3)

        # Check the items are in sequential order.
        self.assertIsInstance(retrieved_items[0], SequencedItem)
        self.assertEqual(retrieved_items[0].sequence_id, item1.sequence_id)
        self.assertEqual(retrieved_items[0].topic, item1.topic)
        self.assertEqual(retrieved_items[0].state, item1.state)

        self.assertIsInstance(retrieved_items[1], SequencedItem)
        self.assertEqual(retrieved_items[1].sequence_id, item3.sequence_id)
        self.assertEqual(retrieved_items[1].topic, item4.topic)
        self.assertEqual(retrieved_items[1].state, item4.state)

        self.assertIsInstance(retrieved_items[2], SequencedItem)
        self.assertEqual(retrieved_items[2].sequence_id, item5.sequence_id)
        self.assertEqual(retrieved_items[2].topic, item5.topic)
        self.assertEqual(retrieved_items[2].state, item5.state)

        # Get items greater than a position.
        retrieved_items = self.record_manager.list_items(sequence_id1, gt=position1)
        self.assertEqual(len(retrieved_items), 2)
        self.assertEqual(retrieved_items[0].state, item4.state)
        self.assertEqual(retrieved_items[1].state, item5.state)

        # Get items greater then or equal to a position.
        retrieved_items = self.record_manager.list_items(sequence_id1, gte=position2)
        self.assertEqual(len(retrieved_items), 2)
        self.assertEqual(retrieved_items[0].state, item4.state)
        self.assertEqual(retrieved_items[1].state, item5.state)

        # Get items less than a position.
        if self.record_manager.can_lt_lte_get_records:
            retrieved_items = self.record_manager.list_items(sequence_id1, lt=position3)
            self.assertEqual(len(retrieved_items), 2)
            self.assertEqual(retrieved_items[0].state, item1.state)
            self.assertEqual(retrieved_items[1].state, item4.state)

        # Get items less then or equal to a position.
        if self.record_manager.can_lt_lte_get_records:
            retrieved_items = self.record_manager.list_items(
                sequence_id1, lte=position2
            )
            self.assertEqual(len(retrieved_items), 2)
            self.assertEqual(retrieved_items[0].state, item1.state)
            self.assertEqual(retrieved_items[1].state, item4.state)

        # Get items greater then or equal to a position and less then or equal to a position.
        if self.record_manager.can_lt_lte_get_records:
            retrieved_items = self.record_manager.list_items(
                sequence_id1, gte=position2, lte=position2
            )
            self.assertEqual(len(retrieved_items), 1)
            self.assertEqual(retrieved_items[0].state, item4.state)

        # Get items greater then or equal to a position and less then a position.
        if self.record_manager.can_lt_lte_get_records:
            retrieved_items = self.record_manager.list_items(
                sequence_id1, gte=position2, lt=position3
            )
            self.assertEqual(len(retrieved_items), 1)
            self.assertEqual(retrieved_items[0].state, item4.state)

        # Get items greater then a position and less then or equal to a position.
        if self.record_manager.can_lt_lte_get_records:
            retrieved_items = self.record_manager.list_items(
                sequence_id1, gt=position1, lte=position2
            )
            self.assertEqual(len(retrieved_items), 1)
            self.assertEqual(retrieved_items[0].state, item4.state)

        # Get items greater a position and less a position.
        if self.record_manager.can_lt_lte_get_records:
            retrieved_items = self.record_manager.list_items(
                sequence_id1, gt=position1, lt=position3
            )
            self.assertEqual(len(retrieved_items), 1)
            self.assertEqual(retrieved_items[0].state, item4.state)

        # Get items with a limit.
        if self.record_manager.can_limit_get_records:
            retrieved_items = self.record_manager.list_items(sequence_id1, limit=1)
            self.assertEqual(len(retrieved_items), 1)
            self.assertEqual(retrieved_items[0].state, item1.state)

        # Get items with a limit, and with descending query (so that we get the last ones).
        if self.record_manager.can_limit_get_records:
            retrieved_items = self.record_manager.list_items(
                sequence_id1, limit=2, query_ascending=False
            )
            self.assertEqual(2, len(retrieved_items))
            self.assertEqual(retrieved_items[0].state, item4.state)
            self.assertEqual(retrieved_items[1].state, item5.state)

        # Get items with a limit and descending query, greater than a position.
        if self.record_manager.can_limit_get_records:
            retrieved_items = self.record_manager.list_items(
                sequence_id1, limit=2, gt=position2, query_ascending=False
            )
            self.assertEqual(1, len(retrieved_items))
            self.assertEqual(retrieved_items[0].state, item5.state)

        # Get items with a limit and descending query, less than a position.
        if self.record_manager.can_limit_get_records:
            retrieved_items = self.record_manager.list_items(
                sequence_id1, limit=2, lt=position3, query_ascending=False
            )
            self.assertEqual(len(retrieved_items), 2)
            self.assertEqual(retrieved_items[0].state, item1.state)
            self.assertEqual(retrieved_items[1].state, item4.state)

        # Get items in descending order, queried in ascending order.
        retrieved_items = self.record_manager.list_items(
            sequence_id1, results_ascending=False
        )
        self.assertEqual(len(retrieved_items), 3)
        self.assertEqual(retrieved_items[0].state, item5.state)
        self.assertEqual(retrieved_items[2].state, item1.state)

        # Get items in descending order, queried in descending order.
        retrieved_items = self.record_manager.list_items(
            sequence_id1, query_ascending=False, results_ascending=False
        )
        self.assertEqual(len(retrieved_items), 3)
        self.assertEqual(retrieved_items[0].state, item5.state)
        self.assertEqual(retrieved_items[2].state, item1.state)

        # Check we can get all the sequence IDs.
        if self.record_manager.can_list_sequence_ids:
            sequence_ids = self.record_manager.list_sequence_ids()
            self.assertEqual(set(sequence_ids), {sequence_id1, sequence_id2})

            # Iterate over all items in all sequences.
            retrieved_items = []
            for sequence_id in sequence_ids:
                retrieved_items += self.record_manager.list_items(sequence_id)

            # Not always in order, but check the number of events.
            self.assertEqual(4, len(retrieved_items))

        if self.record_manager.contiguous_record_ids:
            # Check the record IDs are contiguous.
            records = list(self.record_manager.get_notification_records())
            len_old = len(old_notifications)
            self.assertEqual(len(records), 4 + len_old)
            first = None
            for i, record in enumerate(records):
                if first is None:
                    first = record.id
                self.assertEqual(
                    first + i,
                    record.id,
                    "Woops there's a gap: {}".format([r.id for r in records]),
                )

            # Resume from after the first event.
            retrieved_items = self.record_manager.get_notification_records(
                start=1 + len_old, stop=3 + len_old
            )
            retrieved_items = list(retrieved_items)
            self.assertEqual(len(retrieved_items), 2)
            self.assertEqual(retrieved_items[0].id, 2 + len_old)
            self.assertEqual(retrieved_items[1].id, 3 + len_old)

        # Delete some items.
        if self.record_manager.can_delete_records:
            records = list(self.record_manager.get_records(sequence_id1))
            self.assertTrue(len(records))
            for record in records:
                self.record_manager.delete_record(record)
            records = list(self.record_manager.get_records(sequence_id1))
            self.assertFalse(len(records))

        with self.assertRaises(RecordConflictError):
            self.record_manager.raise_sequenced_item_conflict()

        with self.assertRaises(IndexError):
            self.record_manager.raise_index_error(0)

        with self.assertRaises(RecordConflictError):
            self.record_manager.raise_record_integrity_error(Exception())

        with self.assertRaises(OperationalError):
            self.record_manager.raise_operational_error(Exception())

    @property
    def record_manager(self):
        """
        :rtype: BaseRecordManager
        """
        if self._record_manager is None:
            self._record_manager = self.construct_record_manager()
        return self._record_manager

    @abstractmethod
    def construct_record_manager(self):
        raise NotImplementedError()

    def construct_entity_record_manager(self):
        return self.factory.construct_integer_sequenced_record_manager()

    def construct_snapshot_record_manager(self):
        return self.factory.construct_snapshot_record_manager()

    def construct_timestamp_sequenced_record_manager(self):
        return self.factory.construct_timestamp_sequenced_record_manager()

    @abstractmethod
    def construct_positions(self):
        return None, None, None

    @abstractmethod
    def EXAMPLE_EVENT_TOPIC1(self):
        pass

    @abstractmethod
    def EXAMPLE_EVENT_TOPIC2(self):
        pass


class WithRecordManagers(AbstractDatastoreTestCase):
    """
    Base class for test cases that need a record manager.
    """

    drop_tables = False

    def __init__(self, *args, **kwargs):
        super(WithRecordManagers, self).__init__(*args, **kwargs)
        self._entity_record_manager = None
        self._log_record_manager = None
        self._snapshot_strategy = None

    def setUp(self):
        super(WithRecordManagers, self).setUp()
        if self.datastore:
            self.datastore.setup_connection()
            if self.drop_tables:
                self.datastore.drop_tables()
            self.datastore.setup_tables()

    def tearDown(self):
        self._log_record_manager = None
        self._entity_record_manager = None
        if self._datastore is not None:
            if self.drop_tables:
                self._datastore.drop_tables()
                self._datastore.close_connection()
                self._datastore = None
            else:
                self._datastore.truncate_tables()
        super(WithRecordManagers, self).tearDown()

    @property
    def entity_record_manager(self):
        if self._entity_record_manager is None:
            self._entity_record_manager = self.construct_entity_record_manager()
        return self._entity_record_manager

    @property
    def log_record_manager(self):
        if self._log_record_manager is None:
            self._log_record_manager = self.construct_log_record_manager()
        return self._log_record_manager

    @property
    def snapshot_record_manager(self):
        if self._snapshot_strategy is None:
            self._snapshot_strategy = self.construct_snapshot_record_manager()
        return self._snapshot_strategy

    def construct_entity_record_manager(self):
        return self.factory.construct_integer_sequenced_record_manager()

    def construct_log_record_manager(self):
        return self.factory.construct_timestamp_sequenced_record_manager()

    def construct_snapshot_record_manager(self):
        return self.factory.construct_snapshot_record_manager()


class VersionedEventExample1(EventWithOriginatorVersion, EventWithOriginatorID):
    pass


class VersionedEventExample2(EventWithOriginatorVersion, EventWithOriginatorID):
    pass


class TimestampedEventExample1(EventWithTimestamp, EventWithOriginatorID):
    pass


class TimestampedEventExample2(EventWithTimestamp, EventWithOriginatorID):
    pass


class IntegerSequencedRecordTestCase(RecordManagerTestCase):
    """
    Abstract test case for integer sequenced record managers.
    """

    EXAMPLE_EVENT_TOPIC1 = get_topic(VersionedEventExample1)
    EXAMPLE_EVENT_TOPIC2 = get_topic(VersionedEventExample2)

    def construct_positions(self):
        return 0, 1, 2

    def construct_record_manager(self):
        return self.factory.construct_integer_sequenced_record_manager()


class TimestampSequencedItemTestCase(RecordManagerTestCase):
    """
    Abstract test case for timestamp sequenced record managers.
    """

    EXAMPLE_EVENT_TOPIC1 = get_topic(TimestampedEventExample1)
    EXAMPLE_EVENT_TOPIC2 = get_topic(TimestampedEventExample2)

    def construct_positions(self, num=3):
        while num:
            yield decimaltimestamp()
            num -= 1
            sleep(0.00001)

    def construct_record_manager(self):
        return self.factory.construct_timestamp_sequenced_record_manager()


class AbstractSequencedItemIteratorTestCase(WithRecordManagers):
    """
    Abstract test case for sequenced item iterators.
    """

    ENTITY_ID1 = uuid4()

    def test(self):
        self.setup_sequenced_items()

        assert isinstance(self.entity_record_manager, BaseRecordManager)
        stored_events = self.entity_record_manager.list_items(
            sequence_id=self.entity_id
        )
        stored_events = list(stored_events)
        self.assertEqual(len(stored_events), self.num_events)

        # # Check can get all events in ascending order.
        self.assert_iterator_yields_events(
            is_ascending=True,
            expect_at_start=self.sequenced_items[0].state,
            expect_at_end=self.sequenced_items[-1].state,
            expect_item_count=12,
            expect_page_count=3,
            expect_query_count=3,
            page_size=5,
        )

        # In descending order.
        self.assert_iterator_yields_events(
            is_ascending=False,
            expect_at_start=self.sequenced_items[-1].state,
            expect_at_end=self.sequenced_items[0].state,
            expect_item_count=12,
            expect_page_count=3,
            expect_query_count=3,
            page_size=5,
        )

        # Limit number of items.
        self.assert_iterator_yields_events(
            is_ascending=False,
            expect_at_start=self.sequenced_items[-1].state,
            expect_at_end=self.sequenced_items[-2].state,
            expect_item_count=2,
            expect_page_count=1,
            expect_query_count=1,
            page_size=5,
            limit=2,
        )

        # Match the page size to the number of events.
        self.assert_iterator_yields_events(
            is_ascending=True,
            expect_at_start=self.sequenced_items[0].state,
            expect_at_end=self.sequenced_items[-1].state,
            expect_item_count=12,
            expect_page_count=1,
            expect_query_count=2,
            page_size=self.num_events,
        )

        # Queries are minimised if we set a limit.
        self.assert_iterator_yields_events(
            is_ascending=True,
            expect_at_start=self.sequenced_items[0].state,
            expect_at_end=self.sequenced_items[-1].state,
            expect_item_count=12,
            expect_page_count=1,
            expect_query_count=1,
            page_size=self.num_events,
            limit=12,
        )

    def setup_sequenced_items(self):
        self.sequenced_items = []
        self.number_of_sequenced_items = 12
        for i in range(self.number_of_sequenced_items):
            state = (
                '{"i":%s,"entity_id":"%s","timestamp":%s}'
                % (i, self.entity_id, decimaltimestamp())
            ).encode("utf-8")
            sequenced_item = SequencedItem(
                sequence_id=self.entity_id,
                position=i,
                topic="eventsourcing.example.domain_model#Example.Created",
                state=state,
            )
            self.sequenced_items.append(sequenced_item)
            self.entity_record_manager.record_item(sequenced_item)

    def assert_iterator_yields_events(
        self,
        is_ascending,
        expect_at_start,
        expect_at_end,
        expect_item_count=1,
        expect_page_count=0,
        expect_query_count=0,
        page_size=1,
        limit=None,
    ):
        iterator = self.construct_iterator(is_ascending, page_size, limit=limit)
        retrieved_events = list(iterator)
        self.assertEqual(len(retrieved_events), expect_item_count, retrieved_events)
        self.assertEqual(iterator.page_counter, expect_page_count)
        self.assertEqual(iterator.query_counter, expect_query_count)
        self.assertEqual(iterator.all_item_counter, expect_item_count)
        self.assertEqual(expect_at_start, retrieved_events[0].state)
        self.assertEqual(expect_at_end, retrieved_events[-1].state)

    @property
    def entity_id(self):
        return self.ENTITY_ID1

    @property
    def num_events(self):
        return 12

    def construct_iterator(
        self, is_ascending, page_size, gt=None, lte=None, limit=None
    ):
        return self.iterator_cls(
            record_manager=self.entity_record_manager,
            sequence_id=self.entity_id,
            page_size=page_size,
            gt=gt,
            lte=lte,
            limit=limit,
            is_ascending=is_ascending,
        )

    @property
    @abstractmethod
    def iterator_cls(self):
        """
        Returns iterator class.
        """
        raise NotImplementedError()


class SequencedItemIteratorTestCase(AbstractSequencedItemIteratorTestCase):
    iterator_cls = SequencedItemIterator


class ThreadedSequencedItemIteratorTestCase(AbstractSequencedItemIteratorTestCase):
    iterator_cls = ThreadedSequencedItemIterator


class WithEventPersistence(WithRecordManagers):
    """
    Base class for test cases that need persistence policies.
    """

    def setUp(self):
        super(WithEventPersistence, self).setUp()
        # Setup the persistence subscriber.
        self.entity_event_store = EventStore(
            record_manager=self.entity_record_manager,
            event_mapper=SequencedItemMapper(
                sequenced_item_class=SequencedItem,
                sequence_id_attr_name="originator_id",
                position_attr_name="originator_version",
            ),
        )
        self.log_event_store = EventStore(
            record_manager=self.log_record_manager,
            event_mapper=SequencedItemMapper(
                sequenced_item_class=SequencedItem,
                sequence_id_attr_name="originator_id",
                position_attr_name="timestamp",
            ),
        )
        self.snapshot_store = EventStore(
            record_manager=self.snapshot_record_manager,
            event_mapper=SequencedItemMapper(
                sequenced_item_class=SequencedItem,
                sequence_id_attr_name="originator_id",
                position_attr_name="originator_version",
            ),
        )

        self.integer_sequenced_event_policy = None
        if self.entity_event_store is not None:
            self.integer_sequenced_event_policy = PersistencePolicy(
                event_store=self.entity_event_store,
                persist_event_type=VersionedEntity.Event,
            )

        self.timestamp_sequenced_event_policy = None
        if self.log_event_store is not None:
            self.timestamp_sequenced_event_policy = PersistencePolicy(
                event_store=self.log_event_store, persist_event_type=LoggedEvent
            )

        self.snapshot_policy = None
        if self.snapshot_store is not None:
            self.snapshot_policy = PersistencePolicy(
                event_store=self.snapshot_store, persist_event_type=Snapshot
            )

    def tearDown(self):
        # Close the persistence subscriber.
        if self.snapshot_policy:
            self.snapshot_policy.close()
        if self.timestamp_sequenced_event_policy:
            self.timestamp_sequenced_event_policy.close()
        if self.entity_event_store:
            self.integer_sequenced_event_policy.close()
        super(WithEventPersistence, self).tearDown()



from eventsourcing.domain.model.timebucketedlog import (
    Timebucketedlog,
    TimebucketedlogRepository,
)
from eventsourcing.infrastructure.eventsourcedrepository import EventSourcedRepository


class TimebucketedlogRepo(EventSourcedRepository, TimebucketedlogRepository):
    """
    Event sourced repository for the Example domain model entity.
    """



import datetime
import os
import traceback
from inspect import ismethod
from queue import Empty, Queue
from threading import Event, Lock, Thread
from time import sleep
from typing import Dict, Optional, Tuple, Type

import ray

from eventsourcing.application.process import ProcessApplication
from eventsourcing.application.simple import (
    ApplicationWithConcreteInfrastructure,
    Prompt,
    PromptToPull,
    is_prompt_to_pull,
)
from eventsourcing.domain.model.decorators import retry
from eventsourcing.domain.model.events import subscribe, unsubscribe
from eventsourcing.exceptions import (
    EventSourcingError,
    ExceptionWrapper,
    OperationalError,
    ProgrammingError,
    RecordConflictError,
)
from eventsourcing.infrastructure.base import (
    DEFAULT_PIPELINE_ID,
    RecordManagerWithNotifications,
)
from eventsourcing.system.definition import (
    AbstractSystemRunner,
    System,
    TProcessApplication,
)
from eventsourcing.system.rayhelpers import RayDbJob, RayPrompt
from eventsourcing.system.raysettings import ray_init_kwargs
from eventsourcing.system.runner import DEFAULT_POLL_INTERVAL

ray.init(**ray_init_kwargs)

MAX_QUEUE_SIZE = 1
PAGE_SIZE = 20
MICROSLEEP = 0.000
PROMPT_WITH_NOTIFICATION_IDS = False
PROMPT_WITH_NOTIFICATION_OBJS = False
GREEDY_PULL_NOTIFICATIONS = True


class RayRunner(AbstractSystemRunner):
    """
    Uses actor model framework to run a system of process applications.
    """

    def __init__(
        self,
        system: System,
        pipeline_ids=(DEFAULT_PIPELINE_ID,),
        poll_interval: Optional[int] = None,
        setup_tables: bool = False,
        sleep_for_setup_tables: int = 0,
        db_uri: Optional[str] = None,
        **kwargs
    ):
        super(RayRunner, self).__init__(system=system, **kwargs)
        self.pipeline_ids = list(pipeline_ids)
        self.poll_interval = poll_interval
        self.setup_tables = setup_tables or system.setup_tables
        self.sleep_for_setup_tables = sleep_for_setup_tables
        self.db_uri = db_uri
        self.ray_processes: Dict[Tuple[str, int], RayProcess] = {}

    def start(self):
        """
        Starts all the actors to run a system of process applications.
        """
        # Check we have the infrastructure classes we need.
        for process_class in self.system.process_classes.values():
            if not isinstance(process_class, ApplicationWithConcreteInfrastructure):
                if not self.infrastructure_class:
                    raise ProgrammingError("infrastructure_class is not set")
                elif not issubclass(
                    self.infrastructure_class, ApplicationWithConcreteInfrastructure
                ):
                    raise ProgrammingError(
                        "infrastructure_class is not a subclass of {}".format(
                            ApplicationWithConcreteInfrastructure
                        )
                    )

        # Get the DB_URI.
        # Todo: Support different URI for different application classes.
        env_vars = {}
        db_uri = self.db_uri or os.environ.get("DB_URI")

        if db_uri is not None:
            env_vars["DB_URI"] = db_uri

        # Start processes.
        for pipeline_id in self.pipeline_ids:
            for process_name, process_class in self.system.process_classes.items():
                ray_process_id = RayProcess.remote(
                    application_process_class=process_class,
                    infrastructure_class=self.infrastructure_class,
                    env_vars=env_vars,
                    poll_interval=self.poll_interval,
                    pipeline_id=pipeline_id,
                    setup_tables=self.setup_tables,
                )
                self.ray_processes[(process_name, pipeline_id)] = ray_process_id

        init_ids = []

        for key, ray_process in self.ray_processes.items():
            process_name, pipeline_id = key
            upstream_names = self.system.upstream_names[process_name]
            downstream_names = self.system.downstream_names[process_name]
            downstream_processes = {
                name: self.ray_processes[(name, pipeline_id)]
                for name in downstream_names
            }

            upstream_processes = {}
            for upstream_name in upstream_names:
                upstream_process = self.ray_processes[(upstream_name, pipeline_id)]
                upstream_processes[upstream_name] = upstream_process

            init_ids.append(
                ray_process.init.remote(upstream_processes, downstream_processes)
            )
        ray.get(init_ids)

    def get_ray_process(self, process_name, pipeline_id=DEFAULT_PIPELINE_ID):
        assert isinstance(process_name, str)
        return self.ray_processes[(process_name, pipeline_id)]

    def close(self):
        super(RayRunner, self).close()
        for process in self.ray_processes.values():
            process.stop.remote()

    def get(
        self, process_class: Type[TProcessApplication], pipeline_id=DEFAULT_PIPELINE_ID
    ) -> TProcessApplication:
        assert issubclass(process_class, ProcessApplication)
        process_name = process_class.create_name()
        ray_process = self.get_ray_process(process_name, pipeline_id)
        return ProcessApplicationProxy(ray_process)


@ray.remote
class RayProcess:
    def __init__(
        self,
        application_process_class: Type[ProcessApplication],
        infrastructure_class: Type[ApplicationWithConcreteInfrastructure],
        env_vars: dict = None,
        pipeline_id: int = DEFAULT_PIPELINE_ID,
        poll_interval: int = None,
        setup_tables: bool = False,
    ):
        # Process application args.
        self.application_process_class = application_process_class
        self.infrastructure_class = infrastructure_class
        self.daemon = True
        self.pipeline_id = pipeline_id
        self.poll_interval = poll_interval or DEFAULT_POLL_INTERVAL
        self.setup_tables = setup_tables
        if env_vars is not None:
            os.environ.update(env_vars)

        # Setup threads, queues, and threading events.
        self.readers_lock = Lock()
        self._has_been_prompted = Event()
        self.heads_lock = Lock()
        self.heads = {}
        self.positions_lock = Lock()
        self.positions = {}
        self.positions_initialised = Event()
        self.db_jobs_queue = Queue(maxsize=MAX_QUEUE_SIZE)
        self.upstream_event_queue = Queue(maxsize=MAX_QUEUE_SIZE)
        self.downstream_prompt_queue = Queue()  # no maxsize, call() can put prompt

        self.has_been_stopped = Event()
        self.db_jobs_thread = Thread(target=self.db_jobs)
        self.db_jobs_thread.setDaemon(True)
        self.db_jobs_thread.start()

        self.process_prompts_thread = Thread(target=self._process_prompts)
        self.process_prompts_thread.setDaemon(True)
        self.process_prompts_thread.start()

        self.process_events_thread = Thread(target=self._process_events)
        self.process_events_thread.setDaemon(True)
        self.process_events_thread.start()

        self.push_prompts_thread = Thread(target=self._push_prompts)
        self.push_prompts_thread.setDaemon(True)
        self.push_prompts_thread.start()

        self._notification_rayids = {}
        self._prompted_notifications = {}

    def db_jobs(self):
        # print("Running do_jobs")
        while not self.has_been_stopped.is_set():
            try:
                item = self.db_jobs_queue.get()  # timeout=1)
                self.db_jobs_queue.task_done()
            except Empty:
                if self.has_been_stopped.is_set():
                    break
            else:
                if item is None or self.has_been_stopped.is_set():
                    break
                db_job: RayDbJob = item
                # self.print_timecheck("Doing db job", item)
                try:
                    db_job.execute()
                except Exception as e:
                    if db_job.error is None:
                        print(traceback.format_exc())
                        self._print_timecheck(
                            "Continuing after error running DB job:", e
                        )
                        sleep(1)
                # else:
                #     self.print_timecheck("Done db job", item)

    @retry((OperationalError, RecordConflictError), max_attempts=100, wait=0.01)
    def do_db_job(self, method, args, kwargs):
        db_job = RayDbJob(method, args=args, kwargs=kwargs)
        self.db_jobs_queue.put(db_job)
        db_job.wait()

        if db_job.error:
            raise db_job.error

        # self.print_timecheck("db job delay:", db_job.delay)
        # self.print_timecheck("db job duration:", db_job.duration)

        # self.print_timecheck('db job result:', db_job.result)
        return db_job.result

    def init(self, upstream_processes: dict, downstream_processes: dict) -> None:
        """
        Initialise with actor handles for upstream and downstream processes.

        Need to initialise after construction so that all handles exist.
        """
        self.upstream_processes = upstream_processes
        self.downstream_processes = downstream_processes

        # Subscribe to broadcast prompts published by the process application.
        subscribe(handler=self._enqueue_prompt_to_pull, predicate=is_prompt_to_pull)

        # Construct process application object.
        process_class = self.application_process_class
        if not isinstance(process_class, ApplicationWithConcreteInfrastructure):
            if self.infrastructure_class:
                process_class = process_class.mixin(self.infrastructure_class)
            else:
                raise ProgrammingError("infrastructure_class is not set")

        class MethodWrapper(object):
            def __init__(self, method):
                self.method = method

            def __call__(self, *args, **kwargs):
                try:
                    return self.method(*args, **kwargs)
                except EventSourcingError as e:
                    return ExceptionWrapper(e)

        class ProcessApplicationWrapper(object):
            def __init__(self, process_application):
                self.process_application = process_application

            def __getattr__(self, item):
                attribute = getattr(self.process_application, item)
                if ismethod(attribute):
                    return MethodWrapper(attribute)
                else:
                    return attribute

        def construct_process():
            return process_class(
                pipeline_id=self.pipeline_id, setup_table=self.setup_tables
            )

        process_application = self.do_db_job(construct_process, (), {})
        assert isinstance(process_application, ProcessApplication), process_application
        self.process_wrapper = ProcessApplicationWrapper(process_application)
        self.process_application = process_application

        for upstream_name, ray_notification_log in self.upstream_processes.items():
            # Make the process follow the upstream notification log.
            self.process_application.follow(upstream_name, ray_notification_log)

        self._reset_positions()
        self.positions_initialised.set()

    def _reset_positions(self):
        self.do_db_job(self.__reset_positions, (), {})

    def __reset_positions(self):
        with self.positions_lock:
            for upstream_name in self.upstream_processes:
                recorded_position = self.process_application.get_recorded_position(
                    upstream_name
                )
                self.positions[upstream_name] = recorded_position

    def add_downstream_process(self, downstream_name, ray_process_id):
        self.downstream_processes[downstream_name] = ray_process_id

    def call(self, method_name, *args, **kwargs):
        """
        Method for calling methods on process application object.
        """
        assert self.positions_initialised.is_set(), "Please call .init() first"
        # print("Calling", method_name, args, kwargs)
        if self.process_wrapper:
            method = getattr(self.process_wrapper, method_name)
            return self.do_db_job(method, args, kwargs)
        else:
            raise Exception(
                "Can't call method '%s' before process exists" % method_name
            )

    def prompt(self, prompt: RayPrompt) -> None:
        assert isinstance(prompt, RayPrompt), "Not a RayPrompt: %s" % prompt
        for notification_id, rayid in prompt.notification_ids:
            # self._print_timecheck("Received ray notification ID:", notification_id, rayid)
            self._notification_rayids[(prompt.process_name, notification_id)] = rayid

        latest_head = prompt.head_notification_id
        upstream_name = prompt.process_name
        if PROMPT_WITH_NOTIFICATION_OBJS:
            for notification in prompt.notifications:
                self._prompted_notifications[
                    (upstream_name, notification["id"])
                ] = notification
        if latest_head is not None:
            with self.heads_lock:
                # Update head from prompt.
                if upstream_name in self.heads:
                    if latest_head > self.heads[upstream_name]:
                        self.heads[upstream_name] = latest_head
                        self._has_been_prompted.set()
                else:
                    self.heads[upstream_name] = latest_head
                    self._has_been_prompted.set()
        else:
            self._has_been_prompted.set()

    def _process_prompts(self) -> None:
        # Loop until stop event is set.
        self.positions_initialised.wait()

        while not self.has_been_stopped.is_set():
            try:
                self.__process_prompts()
            except Exception as e:
                if not self.has_been_stopped.is_set():
                    print(traceback.format_exc())
                    print("Continuing after error in 'process prompts' thread:", e)
                    print()
                    sleep(1)

    def __process_prompts(self):
        # Wait until prompted.
        self._has_been_prompted.wait()

        # self.print_timecheck('has been prompted')
        current_heads = {}
        with self.heads_lock:
            self._has_been_prompted.clear()
            for upstream_name in self.upstream_processes.keys():
                current_head = self.heads.get(upstream_name)
                current_heads[upstream_name] = current_head

        for upstream_name in self.upstream_processes.keys():

            with self.positions_lock:
                current_position = self.positions.get(upstream_name)
            first_id = current_position + 1  # request the next one

            current_head = current_heads[upstream_name]
            if current_head is None:
                last_id = None
            elif current_position < current_head:
                if GREEDY_PULL_NOTIFICATIONS:
                    last_id = first_id + PAGE_SIZE - 1
                else:
                    last_id = min(current_head, first_id + PAGE_SIZE - 1)
            else:
                # self.print_timecheck(
                #     "Up to date with", upstream_name, current_position,
                #     current_head
                # )
                continue
                # last_id = first_id + PAGE_SIZE - 1

            # self.print_timecheck(
            #     "Getting notifications in range:",
            #     upstream_name,
            #     "%s -> %s" % (first_id, last_id),
            # )
            upstream_process = self.upstream_processes[upstream_name]
            # Works best without prompted head as last requested,
            # because there might be more notifications since.
            # Todo: However, limit the number to avoid getting too many, and
            #  if we got full quota, then get again.

            notifications = []
            if PROMPT_WITH_NOTIFICATION_IDS or PROMPT_WITH_NOTIFICATION_OBJS:
                if last_id is not None:
                    for notification_id in range(first_id, last_id + 1):
                        if PROMPT_WITH_NOTIFICATION_IDS:
                            try:
                                rayid = self._notification_rayids.pop(
                                    (upstream_name, notification_id)
                                )
                            except KeyError:
                                break
                            else:
                                notification = ray.get(rayid)
                                # self._print_timecheck(
                                #     "Got notification from ray id",
                                #     notification_id,
                                #     rayid,
                                #     notification,
                                # )
                                notifications.append(notification)
                        elif PROMPT_WITH_NOTIFICATION_OBJS:
                            try:
                                notification = self._prompted_notifications.pop(
                                    (upstream_name, notification_id)
                                )
                                # self._print_timecheck(
                                #     "Got notification from prompted notifications dict",
                                #     notification_id,
                                #     notification,
                                # )
                            except KeyError:
                                break
                            else:
                                notifications.append(notification)
                        first_id += 1

            # Pull the ones we don't have.
            if last_id is None or first_id <= last_id:
                # self._print_timecheck("Pulling notifications", first_id, last_id,
                # 'from', upstream_name)
                rayid = upstream_process.get_notifications.remote(first_id, last_id)
                _notifications = ray.get(rayid)
                # self._print_timecheck("Pulled notifications", _notifications)
                notifications += _notifications

            # self.print_timecheck(
            #     "Obtained notifications:", len(notifications), 'from',
            #     upstream_name
            # )

            if len(notifications):

                if len(notifications) == PAGE_SIZE:
                    # self._print_timecheck("Range limit reached, reprompting...")
                    self._has_been_prompted.set()

                position = notifications[-1]["id"]
                with self.positions_lock:
                    current_position = self.positions[upstream_name]
                    if current_position is None or position > current_position:
                        self.positions[upstream_name] = position

            queue_item = []
            for notification in notifications:
                # Check causal dependencies.
                self.process_application.check_causal_dependencies(
                    upstream_name, notification.get("causal_dependencies")
                )
                # Get domain event from notification.
                event = self.process_application.get_event_from_notification(
                    notification
                )
                # self.print_timecheck("obtained event", event)

                # Put domain event on the queue, for event processing.
                queue_item.append((event, notification["id"], upstream_name))
            self.upstream_event_queue.put(queue_item)
            sleep(MICROSLEEP)

    def get_notifications(self, first_notification_id, last_notification_id):
        """
        Returns a list of notifications, with IDs from first_notification_id
        to last_notification_id, inclusive. IDs are 1-based sequence.

        This is called by the "process prompts" thread of a downstream process.
        """
        return self.do_db_job(
            self._get_notifications, (first_notification_id, last_notification_id), {}
        )

    def _get_notifications(self, first_notification_id, last_notification_id):
        record_manager = self.process_application.event_store.record_manager
        assert isinstance(record_manager, RecordManagerWithNotifications)
        start = first_notification_id - 1
        stop = last_notification_id
        return list(record_manager.get_notifications(start, stop))

    def _process_events(self):
        while not self.has_been_stopped.is_set():
            try:
                self.__process_events()
            except Exception as e:
                print(traceback.format_exc())
                print("Continuing after error in 'process events' thread:", e)
                sleep(1)

    def __process_events(self):
        try:
            queue_item = self.upstream_event_queue.get()  # timeout=5)
            self.upstream_event_queue.task_done()
        except Empty:
            if self.has_been_stopped.is_set():
                return
        else:
            if queue_item is None or self.has_been_stopped.is_set():
                return

            for (domain_event, notification_id, upstream_name) in queue_item:
                # print("Processing upstream event:", (domain_event,
                # notification_id, upstream_name))

                new_events, new_records = (), ()
                while not self.has_been_stopped.is_set():
                    try:
                        new_events, new_records = self.do_db_job(
                            method=self.process_application.process_upstream_event,
                            args=(domain_event, notification_id, upstream_name),
                            kwargs={},
                        )
                        break
                    except Exception as e:
                        print(traceback.format_exc())
                        self._print_timecheck(
                            "Retrying to reprocess event after error:", e
                        )
                        sleep(1)
                        # Todo: Forever? What if this is the wrong event?

                if self.has_been_stopped.is_set():
                    return

                # if new_events:
                #     self._print_timecheck("new events", len(new_events), new_events)

                notifications = ()
                notification_ids = ()
                notifiable_events = [e for e in new_events if e.__notifiable__]
                if len(notifiable_events):
                    if PROMPT_WITH_NOTIFICATION_IDS or PROMPT_WITH_NOTIFICATION_OBJS:
                        manager = self.process_application.event_store.record_manager
                        assert isinstance(manager, RecordManagerWithNotifications)
                        notification_id_name = manager.notification_id_name
                        notifications = []
                        for record in new_records:
                            if isinstance(
                                getattr(record, notification_id_name, None), int
                            ):
                                notifications.append(
                                    manager.create_notification_from_record(record)
                                )
                        if len(notifications):
                            head_notification_id = notifications[-1]["id"]
                            if PROMPT_WITH_NOTIFICATION_IDS:
                                notification_ids = self._put_notifications_in_ray_object_store(
                                    notifications
                                )
                                # Clear the notifications, avoid sending with IDs.
                                notifications = ()
                        else:
                            head_notification_id = self._get_max_notification_id()
                    else:
                        head_notification_id = self._get_max_notification_id()

                    prompt = RayPrompt(
                        self.process_application.name,
                        self.process_application.pipeline_id,
                        head_notification_id,
                        notification_ids,
                        notifications,
                    )

                    # self.print_timecheck(
                    #     "putting prompt on downstream " "prompt queue",
                    #     self.downstream_prompt_queue.qsize(),
                    # )
                    self.downstream_prompt_queue.put(prompt)
                    sleep(MICROSLEEP)
                    # self.print_timecheck(
                    #     "put prompt on downstream prompt " "queue"
                    # )
        # sleep(0.1)

    def _put_notifications_in_ray_object_store(self, notifications):
        notification_ids = [(n["id"], ray.put(n)) for n in notifications]
        return notification_ids

    def _enqueue_prompt_to_pull(self, prompt):
        # print("Enqueing locally published prompt:", prompt)
        self.downstream_prompt_queue.put(prompt)
        sleep(MICROSLEEP)

    def _push_prompts(self) -> None:
        while not self.has_been_stopped.is_set():
            try:
                self.__push_prompts()
            except Exception as e:
                print(traceback.format_exc())
                print("Continuing after error in 'push prompts' thread:", e)
                sleep(1)

    def __push_prompts(self):
        try:
            item = self.downstream_prompt_queue.get()  # timeout=1)
            self.downstream_prompt_queue.task_done()
            # Todo: Instead, drain the queue and consolidate prompts.
        except Empty:
            self._print_timecheck(
                "timed out getting item from downstream prompt " "queue"
            )
            if self.has_been_stopped.is_set():
                return
        else:
            # self.print_timecheck("task done on downstream prompt queue")
            if item is None or self.has_been_stopped.is_set():
                return
            elif isinstance(item, PromptToPull):
                if item.head_notification_id:
                    head_notification_id = item.head_notification_id
                else:
                    head_notification_id = self._get_max_notification_id()
                prompt = RayPrompt(
                    self.process_application.name,
                    self.process_application.pipeline_id,
                    head_notification_id,
                )
            else:
                prompt = item
            # self._print_timecheck('pushing prompt with', prompt.notification_ids)
            prompt_response_ids = []
            # self.print_timecheck("pushing prompts", prompt)
            for downstream_name, ray_process in self.downstream_processes.items():
                prompt_response_ids.append(ray_process.prompt.remote(prompt))
                if self.has_been_stopped.is_set():
                    return
                # self._print_timecheck("pushed prompt to", downstream_name)
            ray.get(prompt_response_ids)
            # self._print_timecheck("pushed prompts")

    def _get_max_notification_id(self):
        """
        Returns the highest notification ID of this process application.
        :return:
        """
        record_manager = self.process_application.event_store.record_manager
        assert isinstance(record_manager, RecordManagerWithNotifications)
        max_notification_id = self.do_db_job(
            record_manager.get_max_notification_id, (), {}
        )
        # self.print_timecheck("MAX NOTIFICATION ID in DB:", max_notification_id)
        return max_notification_id

    def stop(self):
        """
        Stops the process.
        """
        self.has_been_stopped.set()
        self.process_application.close()
        unsubscribe(handler=self._enqueue_prompt_to_pull, predicate=is_prompt_to_pull)

    def _print_timecheck(self, activity, *args):
        # pass
        process_name = self.application_process_class.__name__.lower()
        print(
            "Timecheck",
            datetime.datetime.now(),
            self.pipeline_id,
            process_name,
            activity,
            *args
        )


class ProcessApplicationProxy:
    def __init__(self, ray_process: RayProcess):
        self.ray_process: RayProcess = ray_process

    def __getattr__(self, item):
        return AttributeProxy(self.ray_process, item)


class AttributeProxy:
    def __init__(self, ray_process: RayProcess, attribute_name: str):
        self.ray_process: RayProcess = ray_process
        self.attribute_name = attribute_name

    def __call__(self, *args, **kwargs):
        ray_id = self.ray_process.call.remote(self.attribute_name, *args, **kwargs)
        return_value = ray.get(ray_id)
        if isinstance(return_value, ExceptionWrapper):
            raise return_value.e
        else:
            return return_value



from typing import Iterable, NamedTuple, Optional
from uuid import UUID

from eventsourcing.exceptions import ConcurrencyError, RecordConflictError
from eventsourcing.infrastructure.base import AbstractEventStore, TRecordManager
from eventsourcing.infrastructure.iterators import SequencedItemIterator
from eventsourcing.whitehead import TEvent

# Todo: Unify iterators in EventStore and in NotificationLog,
#  by pushing behaviour down to record manager?


class EventStore(AbstractEventStore[TEvent, TRecordManager]):
    """
    Event store appends domain events to stored sequences. It uses
    a record manager to map named tuples to database
    records, and it uses a sequenced item mapper to map named
    tuples to application-level objects.
    """

    iterator_class = SequencedItemIterator

    def store_events(self, events: Iterable[TEvent]) -> None:
        """
        Appends given domain event, or list of domain events, to their sequence.

        :param events: domain event, or list of domain events
        """

        # Convert to sequenced item.
        sequenced_items = self.items_from_events(events)

        # Append to the sequenced item(s) to the sequence.
        try:
            self.record_manager.record_items(sequenced_items)
        except RecordConflictError as e:
            raise ConcurrencyError(e)

    def items_from_events(self, events: Iterable[TEvent]) -> Iterable[NamedTuple]:
        """
        Maps domain event to sequenced item namedtuple.

        An iterable of events.
        """
        # Convert the domain event(s) to sequenced item(s).
        return map(self.event_mapper.item_from_event, events)

    def iter_events(
        self,
        originator_id: UUID,
        gt: Optional[int] = None,
        gte: Optional[int] = None,
        lt: Optional[int] = None,
        lte: Optional[int] = None,
        limit: Optional[int] = None,
        is_ascending: bool = True,
        page_size: Optional[int] = None,
    ) -> Iterable[TEvent]:
        """
        Gets domain events from the sequence identified by `originator_id`.

        :param originator_id: ID of a sequence of events
        :param gt: get items after this position
        :param gte: get items at or after this position
        :param lt: get items before this position
        :param lte: get items before or at this position
        :param limit: get limited number of items
        :param is_ascending: get items from lowest position
        :param page_size: restrict and repeat database query
        :return: list of domain events
        """
        if self.record_manager.can_limit_get_records and page_size:
            sequenced_items: Iterable = self.iterator_class(
                record_manager=self.record_manager,
                sequence_id=originator_id,
                page_size=page_size,
                gt=gt,
                gte=gte,
                lt=lt,
                lte=lte,
                limit=limit,
                is_ascending=is_ascending,
            )
        else:
            sequenced_items = self.record_manager.get_items(
                sequence_id=originator_id,
                gt=gt,
                gte=gte,
                lt=lt,
                lte=lte,
                limit=limit,
                query_ascending=is_ascending,
                results_ascending=is_ascending,
            )

        # Deserialize to domain events.
        return map(self.event_mapper.event_from_item, sequenced_items)

    def get_event(self, originator_id: UUID, position: int) -> TEvent:
        """
        Gets a domain event from the sequence identified by `originator_id`
        at position `eq`.

        :param originator_id: ID of a sequence of events
        :param position: get item at this position
        :return: domain event
        """
        sequenced_item = self.record_manager.get_item(
            sequence_id=originator_id, position=position
        )
        return self.event_mapper.event_from_item(sequenced_item)

    def get_most_recent_event(
        self, originator_id: UUID, lt: Optional[int] = None, lte: Optional[int] = None
    ) -> Optional[TEvent]:
        """
        Gets a domain event from the sequence identified by `originator_id`
        at the highest position.

        :param originator_id: ID of a sequence of events
        :param lt: get highest before this position
        :param lte: get highest at or before this position
        :return: domain event
        """
        events = self.list_events(
            originator_id=originator_id, lt=lt, lte=lte, limit=1, is_ascending=False
        )
        len_events = len(events)
        if len_events == 1:
            return events[0]
        elif len_events == 0:
            return None
        else:
            raise AssertionError("Too many events: %s" % len_events)

    def all_events(self) -> Iterable[TEvent]:
        """
        Yields all domain events in the event store.

        This method iterates over the sequence IDs,
        and returns all the events for each sequence
        effectively concatenated together.

        Use a notification log to propagate events from
        an application as a stable append-only sequence.
        """
        for originator_id in self.record_manager.all_sequence_ids():
            for domain_event in self.iter_events(
                originator_id=originator_id, page_size=100
            ):
                yield domain_event
